"""
Integration tests for data synchronization.
"""

from flask import current_app

from kerko.storage import SearchIndexError
from kerko.sync import zotero
from kerko.sync.cache import sync_cache
from kerko.sync.index import sync_index
from tests.integration_testing import EmptyLibraryTestCase, PopulatedLibraryTestCase


class SyncPopulatedLibraryTestCase(PopulatedLibraryTestCase):
    """
    Test data synchronization with a populated Zotero library.

    If this fails, all other integration tests are likely to fail as well.
    """

    def test_sync_library_context(self):
        zotero_credentials = zotero.init_zotero()
        library_context = zotero.request_library_context(zotero_credentials)
        self.assertEqual(
            library_context.library_id,
            current_app.config["ZOTERO_LIBRARY_ID"],
        )
        self.assertEqual(
            library_context.library_type,
            current_app.config["ZOTERO_LIBRARY_TYPE"],
        )
        self.assertEqual(set(library_context.item_types.keys()), set(self.ZOTERO_ITEM_TYPES))
        self.assertEqual(set(library_context.item_fields.keys()), set(self.ZOTERO_ITEM_TYPES))
        self.assertEqual(set(library_context.creator_types.keys()), set(self.ZOTERO_ITEM_TYPES))
        # TODO: Assert more things.

    def test_sync_index(self):
        self.assertGreater(sync_cache(), 0, "Cache is empty!")
        self.assertGreater(sync_index(), 0, "Index is empty!")
        # TODO: Assert more things.


class SyncEmptyLibraryTestCase(EmptyLibraryTestCase):
    """
    Test data synchronization with an empty Zotero library.

    If this fails, other integration tests are likely to fail as well.
    """

    def test_sync_index(self):
        self.assertEqual(sync_cache(), 0)
        self.assertRaises(SearchIndexError, sync_index)



"""
Processors for encoding/decoding values and labels to/from the search index.

Encoding refers to preparing a value for use in the search index, and decoding
to extracting a value and label for use in the search interface.
"""

import datetime
import json
from abc import ABC, abstractmethod

from flask_babel import lazy_gettext as _


class BaseFieldCodec(ABC):
    """
    Encode or decode a value.
    """

    @abstractmethod
    def encode(self, value):
        pass

    @abstractmethod
    def decode(self, encoded_value):
        pass


class IdentityFieldCodec(BaseFieldCodec):
    """
    Encode or decode a value without any transformation.
    """

    def encode(self, value):
        return value

    def decode(self, encoded_value):
        return encoded_value


class JSONFieldCodec(BaseFieldCodec):
    """
    Encode or decode a value to/from a JSON string.
    """

    def encode(self, value):
        return json.dumps(value)

    def decode(self, encoded_value):
        return json.loads(encoded_value)


class BaseFacetCodec:
    """
    Encode or decode values and labels.
    """

    def encode(self, value):
        """
        Encode the given value.

        This default implementation returns the value as-is.
        """
        return value

    def decode(self, encoded_value, default_value=None, default_label=None):  # noqa: ARG002
        """
        Decode the given value into its original value and label components.

        Always pass encoded value as the default_value, unless there is a more
        reasonable default. Pass None only when None is really wanted for
        values that cannot be decoded. Some codecs may ignore those defaults.

        This default implementation returns the encoded value as-is for both
        its value and its label.

        :return: A (value, label) tuple.
        """
        return encoded_value, encoded_value

    def transform_for_query(self, value):
        """
        Transform a value before use in a query.

        This default implementation returns the encoded value as-is.
        """
        return value


class BooleanFacetCodec(BaseFacetCodec):
    def __init__(self, true_value="t", false_value="f", true_label=_("yes"), false_label=_("no")):
        self.true_value = true_value
        self.false_value = false_value
        self.true_label = true_label
        self.false_label = false_label

    def decode(self, encoded_value, default_value=None, default_label=None):  # noqa: ARG002
        # Note: explicit encode not necessary as Whoosh automatically encodes
        # booleans as 't' or 'f' values. However the value returned by Whoosh
        # is not consistent whether it comes from group (faceting) results
        # (where it is returned as 't' or 'f') or stored document fields (where
        # it is returned as a bool). Therefore, we also check for bool values.
        if isinstance(encoded_value, bool):
            encoded_value = self.true_value if encoded_value else self.false_value
        return encoded_value, (
            self.true_label if encoded_value == self.true_value else self.false_label
        )


class LabelFacetCodec(BaseFacetCodec):
    """Decode a value suffixed with its label."""

    def __init__(self, label_separator=":", **kwargs):
        super().__init__(**kwargs)
        self.label_separator = label_separator

    def encode(self, value):
        """
        Encode a (value, label) tuple into a proper string for faceting.

        The label is not needed for facet queries, but is appended to the value
        for display purposes, to avoid extra queries at search time for getting
        those labels.

        :param value: A (value, label) tuple.
        """
        value, label = value
        return value + self.label_separator + label

    def decode(self, encoded_value, default_value=None, default_label=None):
        if not encoded_value:
            return default_value, default_label
        elif self.label_separator in encoded_value:
            value, label = encoded_value.split(self.label_separator, 1)
            return value, label
        return encoded_value, encoded_value

    def transform_for_query(self, value):
        """
        Transform a value before use in a query.

        This assumes a Prefix query is being performed.

        Encoded item type values are composed of the query value, a separator,
        and the label. Queries normally search by prefix to search the value and
        ignore the label. However, some values may be a prefix of others (e.g.
        "book" is a prefix of "bookSection"). To avoid these ambiguities, this
        method transforms the query value to include the separator (e.g. "book:"
        and "bookSection:").
        """
        return value + self.label_separator


class CollectionFacetCodec(LabelFacetCodec):
    def __init__(self, path_separator=".", **kwargs):
        kwargs.setdefault("label_separator", " ")
        super().__init__(**kwargs)
        self.path_separator = path_separator

    def encode(self, value):
        """
        Encode a collection path into a proper string for faceting.

        :param value: A (path, label) tuple.

        If the specified path is ['id1', 'id2', 'id3'], the encoded value will
        have the form: "id1.id2.id3 label".
        """
        path, label = value
        encoded_value = self.path_separator.join(path)
        if label:
            encoded_value += self.label_separator + label
        return encoded_value


class ItemTypeFacetCodec(LabelFacetCodec):
    def __init__(self, **kwargs):
        kwargs.setdefault("label_separator", ":")
        super().__init__(**kwargs)


class YearTreeFacetCodec(BaseFacetCodec):
    def __init__(self, path_separator=".", **kwargs):
        super().__init__(**kwargs)
        self.path_separator = path_separator

    def encode(self, value):
        """Encode a year tree path into a proper string for faceting."""
        return self.path_separator.join(value)

    def decode(self, encoded_value, default_value=None, default_label=None):
        if not encoded_value:
            return default_value, default_label
        path = encoded_value.split(self.path_separator)
        assert len(path) in [1, 2, 3]

        if len(path) == 3:  # noqa: PLR2004
            return encoded_value, path[-1]

        if len(path) == 1:
            # Century.
            start = path[0]
            end = int(path[0]) + 99
        else:  # len(path) == 2:
            # Decade.
            start = path[1]
            end = int(path[1]) + 9
        if end >= datetime.datetime.now().year:
            end = datetime.datetime.now().year
            if start == end:
                return encoded_value, _("In {}").format(end)
        return encoded_value, _("Between {} and {}").format(start, end)



"""
Utilities for integration testing.

This allows testing the full sync process, with mock Zotero API responses.
"""

import pathlib
import re
import tempfile
import unittest

import responses
from flask import Flask
from flask_babel import Babel, Domain
from flask_bootstrap import Bootstrap4

import kerko
from kerko import extractors, transformers
from kerko.composer import Composer
from kerko.config_helpers import config_update, parse_config
from kerko.storage import delete_storage
from kerko.sync.cache import sync_cache
from kerko.sync.index import sync_index


class MockLibraryTestCase(unittest.TestCase):
    """Base test case providing mock Zotero API responses."""

    URL_PREFIX = "/bibliography"

    ZOTERO_RESPONSE_HEADERS = {
        "Zotero-API-Version": "3",
        "Zotero-Schema-Version": "15",
    }

    ZOTERO_ITEM_TYPES = [
        "artwork",
        "audioRecording",
        "bill",
        "blogPost",
        "book",
        "bookSection",
        "case",
        "computerProgram",
        "conferencePaper",
        "dictionaryEntry",
        "document",
        "email",
        "encyclopediaArticle",
        "film",
        "forumPost",
        "hearing",
        "instantMessage",
        "interview",
        "journalArticle",
        "letter",
        "magazineArticle",
        "manuscript",
        "map",
        "newspaperArticle",
        "note",
        "patent",
        "podcast",
        "preprint",
        "presentation",
        "radioBroadcast",
        "report",
        "statute",
        "tvBroadcast",
        "thesis",
        "videoRecording",
        "webpage",
    ]

    @staticmethod
    def get_response(response_name):
        response_path = pathlib.Path(__file__).parent / "api_responses" / (response_name + ".json")
        with response_path.open() as f:
            return f.read()

    @classmethod
    def init_blueprints(cls):
        cls.app.register_blueprint(kerko.make_blueprint(), url_prefix=cls.URL_PREFIX)

    @classmethod
    def init_extensions(cls):
        cls._babel_domain = Domain()
        cls._babel = Babel(default_domain=cls._babel_domain)
        cls._babel.init_app(cls.app)
        cls._bootstrap = Bootstrap4()
        cls._bootstrap.init_app(cls.app)

    @classmethod
    def init_config(cls):
        config_update(cls.app.config, kerko.DEFAULTS)
        cls.app.config["SECRET_KEY"] = "not-so-secret-secret"
        cls.app.config["ZOTERO_API_KEY"] = "xxxxxxxxxxxxxxxxxxxxxxxx"
        cls.app.config["ZOTERO_LIBRARY_ID"] = "9999999"
        cls.app.config["ZOTERO_LIBRARY_TYPE"] = "group"
        cls.app.config["DATA_PATH"] = cls.temp_dir.name
        parse_config(cls.app.config)
        cls.app.config["kerko_composer"] = Composer(cls.app.config)

        # Add alternate_id to help retrieving and testing specific items.
        cls.app.config["kerko_composer"].fields["alternate_id"].extractor.extractors.append(
            extractors.TransformerExtractor(
                extractor=extractors.ItemDataExtractor(key="extra"),
                transformers=[
                    transformers.find(
                        regex=r"^\s*KerkoTestID\s*:\s*([0-9\-A-Z]+)\s*$",
                        flags=re.IGNORECASE | re.MULTILINE,
                        max_matches=1,
                    ),
                ],
            )
        )

    @classmethod
    def add_responses(cls):
        """
        Add mock API responses.

        This class only implements those responses that are always the same
        regardless of the Zotero library.

        Subclasses should override this method to add more responses.
        """
        cls.responses.add(
            responses.GET,
            "https://api.zotero.org/itemTypes",
            body=cls.get_response("itemTypes"),
            content_type="application/json",
            headers=cls.ZOTERO_RESPONSE_HEADERS,
        )
        for item_type in cls.ZOTERO_ITEM_TYPES:
            cls.responses.add(
                responses.GET,
                re.compile(
                    re.escape(
                        f"https://api.zotero.org/itemTypeFields?itemType={item_type}&locale=en-US"
                    )
                    + r"(\&timeout=[0-9]+)?"
                ),
                body=cls.get_response(f"itemTypeFields_{item_type}"),
                content_type="application/json",
                headers=cls.ZOTERO_RESPONSE_HEADERS,
            )
            cls.responses.add(
                responses.GET,
                re.compile(
                    re.escape(
                        f"https://api.zotero.org/itemTypeCreatorTypes?itemType={item_type}&locale=en-US"
                    )
                    + r"(\&timeout=[0-9]+)?"
                ),
                body=cls.get_response(f"itemTypeCreatorTypes_{item_type}"),
                content_type="application/json",
                headers=cls.ZOTERO_RESPONSE_HEADERS,
            )

    @classmethod
    def setUpClass(cls):
        """
        Prepare the test fixtures.

        These are set up at the class level helps avoid repeating the relatively
        slow Kerko sync process for every test method.
        """
        cls.app = Flask(__name__)
        cls.temp_dir = tempfile.TemporaryDirectory(prefix="kerko-tests-")
        cls.init_config()
        cls.init_blueprints()
        cls.init_extensions()
        ctx = cls.app.app_context()
        ctx.push()

        # Setup mock responses, per https://github.com/getsentry/responses#responses-inside-a-unittest-setup.
        cls.responses = responses.RequestsMock()
        cls.responses.start()

        cls.add_responses()

        cls.addClassCleanup(cls.responses.stop)
        cls.addClassCleanup(cls.responses.reset)

        # Make sure the data directory is empty before synchronizing.
        delete_storage("cache")
        delete_storage("index")

    @classmethod
    def tearDownClass(cls):
        cls.temp_dir.cleanup()


class PopulatedLibraryTestCase(MockLibraryTestCase):
    """Test case providing mock Zotero API responses for a populated library."""

    ZOTERO_ITEMS_TOTAL_RESULTS = "10"
    ZOTERO_ITEMS_LAST_MODIFIED_VERSION = "26"

    @classmethod
    def add_responses(cls):
        super().add_responses()
        cls.responses.add(
            responses.GET,
            "https://api.zotero.org/groups/9999999/collections?start=0&limit=100&format=json",
            body=cls.get_response("collections"),
            content_type="application/json",
            headers=cls.ZOTERO_RESPONSE_HEADERS,
        )
        # Fallback for other 'collections' requests.
        cls.responses.add(
            responses.GET,
            "https://api.zotero.org/groups/9999999/collections",
            body="[]",
            content_type="application/json",
            headers=cls.ZOTERO_RESPONSE_HEADERS,
        )
        cls.responses.add(
            responses.GET,
            "https://api.zotero.org/groups/9999999/items?since=0&start=0&limit=100&sort=dateAdded&direction=asc&include=bib%2Cbibtex%2Ccoins%2Cdata%2Cris&style=apa&format=json",
            body=cls.get_response("items"),
            content_type="application/json",
            headers={
                **cls.ZOTERO_RESPONSE_HEADERS,
                **{
                    "Total-Results": cls.ZOTERO_ITEMS_TOTAL_RESULTS,
                    "Last-Modified-Version": cls.ZOTERO_ITEMS_LAST_MODIFIED_VERSION,
                },
            },
        )
        cls.responses.add(
            responses.GET,
            f"https://api.zotero.org/groups/9999999/items?since=0&start={cls.ZOTERO_ITEMS_TOTAL_RESULTS}&limit=100&sort=dateAdded&direction=asc&include=bib%2Cbibtex%2Ccoins%2Cdata%2Cris&style=apa&format=json",
            body="[]",
            content_type="application/json",
            headers={
                **cls.ZOTERO_RESPONSE_HEADERS,
                **{
                    "Total-Results": cls.ZOTERO_ITEMS_TOTAL_RESULTS,
                },
            },
        )
        cls.responses.add(
            responses.GET,
            "https://api.zotero.org/groups/9999999/items?limit=1&format=json",
            body=cls.get_response("items_versions"),
            content_type="application/json",
            headers={
                **cls.ZOTERO_RESPONSE_HEADERS,
                **{
                    "Total-Results": cls.ZOTERO_ITEMS_TOTAL_RESULTS,
                    "Last-Modified-Version": cls.ZOTERO_ITEMS_LAST_MODIFIED_VERSION,
                },
            },
        )
        cls.responses.add(
            responses.GET,
            "https://api.zotero.org/groups/9999999/fulltext?since=0",
            body=cls.get_response("fulltext"),
            content_type="application/json",
            headers=cls.ZOTERO_RESPONSE_HEADERS,
        )


class EmptyLibraryTestCase(MockLibraryTestCase):
    """Test case providing mock Zotero API responses for an empty library."""

    @classmethod
    def add_responses(cls):
        super().add_responses()
        # Response for all 'collections' requests.
        cls.responses.add(
            responses.GET,
            "https://api.zotero.org/groups/9999999/collections",
            body="[]",  # No collections.
            content_type="application/json",
            headers=cls.ZOTERO_RESPONSE_HEADERS,
        )
        # Response for all 'items' requests.
        cls.responses.add(
            responses.GET,
            "https://api.zotero.org/groups/9999999/items",
            body="[]",  # No items.
            content_type="application/json",
            headers={
                **cls.ZOTERO_RESPONSE_HEADERS,
                **{
                    "Total-Results": "0",
                    "Last-Modified-Version": "1",
                },
            },
        )
        # Response for all 'fulltext' requests.
        cls.responses.add(
            responses.GET,
            "https://api.zotero.org/groups/9999999/fulltext",
            body="{}",
            content_type="application/json",
            headers=cls.ZOTERO_RESPONSE_HEADERS,
        )


class SynchronizedTestCase(PopulatedLibraryTestCase):
    """
    Test case providing synchronized data from the integration testing library.

    If tests based on this test case fail, check tests from `test_sync`, and fix
    those first if they are also failing.
    """

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        sync_cache()
        sync_index()



import pathlib

from hatchling.builders.hooks.plugin.interface import BuildHookInterface


class CustomBuildHook(BuildHookInterface):
    def initialize(self, version, build_data):  # noqa: ARG002
        from babel.messages.frontend import compile_catalog

        cmd = compile_catalog()
        cmd.directory = pathlib.Path("src") / "kerko" / "translations"
        cmd.domain = "kerko"
        cmd.statistics = True
        cmd.finalize_options()
        cmd.run()



"""
Tests for the XML sitemap views.
"""

from flask import url_for
from lxml import etree

from kerko.searcher import Searcher
from kerko.storage import open_index
from tests.integration_testing import SynchronizedTestCase


class PopulatedSitemapTestCase(SynchronizedTestCase):
    """Test the sitemap with a populated library."""

    def setUp(self):
        self.namespaces = {
            "ns": "http://www.sitemaps.org/schemas/sitemap/0.9",
        }

    def test_sitemap_index(self):
        """Test the XML sitemap index."""

        with self.app.test_client() as client:
            response = client.get(f"{self.URL_PREFIX}/sitemap.xml")
            self.assertEqual(response.status_code, 200)

            root = etree.fromstring(response.get_data(as_text=True).encode("utf-8"))
            self.assertEqual(
                root.xpath("/ns:sitemapindex/ns:sitemap/ns:loc/text()", namespaces=self.namespaces),
                [url_for("kerko.sitemap", page_num=1, _external=True)],
            )

    def test_sitemap(self):
        """Test the XML sitemap."""

        with self.app.test_client() as client:
            response = client.get(f"{self.URL_PREFIX}/sitemap1.xml")
            self.assertEqual(response.status_code, 200)

            root = etree.fromstring(response.get_data(as_text=True).encode("utf-8"))
            urls = root.xpath("/ns:urlset/ns:url/ns:loc/text()", namespaces=self.namespaces)
            self.assertEqual(
                len(set(urls)),
                len(urls),
                "One or more sitemap URLs are not unique",
            )
            self.assertEqual(
                len(urls),
                len(Searcher(open_index("index")).search()),
                "The number of URLs in the sitemap does not match the number of items.",
            )
            for url in urls:
                self.assertEqual(client.get(url).status_code, 200)

    def test_sitemap_out_of_range(self):
        """Test out of range sitemaps."""

        with self.app.test_client() as client:
            response = client.get(f"{self.URL_PREFIX}/sitemap0.xml")
            self.assertEqual(response.status_code, 404)

        with self.app.test_client() as client:
            response = client.get(f"{self.URL_PREFIX}/sitemap500.xml")
            self.assertEqual(response.status_code, 404)

        with self.app.test_client() as client:
            response = client.get(f"{self.URL_PREFIX}/sitemapbomb.xml")
            self.assertEqual(response.status_code, 404)



"""
Unit tests for the pager module.
"""

import unittest

from flask import Flask

import kerko
from kerko.config_helpers import config_update
from kerko.shortcuts import config
from kerko.views import pager


class SectionsTestCase(unittest.TestCase):
    """Test pager sections when there is a single page."""

    def setUp(self):
        self.app = Flask(__name__)
        config_update(self.app.config, kerko.DEFAULTS)
        self.app.register_blueprint(kerko.make_blueprint(), url_prefix="/bibliography")
        ctx = self.app.app_context()
        ctx.push()

    def test_default_config(self):
        # This setting impacts the results of pager functions.
        self.assertEqual(config("kerko.pagination.pager_links"), 4)

    def test_on_page_1_of_1(self):
        sections = pager.get_sections(page_num=1, page_count=1)
        self.assertIsNone(sections)
        pages = pager.get_page_numbers(sections)
        self.assertFalse(pages)

    def test_on_page_1_of_2(self):
        sections = pager.get_sections(page_num=1, page_count=2)
        self.assertNotIn("previous", sections)
        self.assertNotIn("first", sections)
        self.assertFalse(sections["before"])
        self.assertEqual(sections["current"], [1])
        self.assertEqual(sections["after"], [2])
        self.assertEqual(sections["last"], [2])
        self.assertEqual(sections["next"], [2])
        pages = pager.get_page_numbers(sections)
        self.assertEqual(pages, [1, 2])

    def test_on_page_2_of_2(self):
        sections = pager.get_sections(page_num=2, page_count=2)
        self.assertEqual(sections["previous"], [1])
        self.assertEqual(sections["first"], [1])
        self.assertEqual(sections["before"], [1])
        self.assertEqual(sections["current"], [2])
        self.assertFalse(sections["after"])
        self.assertNotIn("last", sections)
        self.assertNotIn("next", sections)
        pages = pager.get_page_numbers(sections)
        self.assertEqual(pages, [1, 2])

    def test_on_page_3_of_5(self):
        sections = pager.get_sections(page_num=3, page_count=5)
        self.assertEqual(sections["previous"], [2])
        self.assertEqual(sections["first"], [1])
        self.assertEqual(sections["before"], [1, 2])
        self.assertEqual(sections["current"], [3])
        self.assertEqual(sections["after"], [4, 5])
        self.assertEqual(sections["last"], [5])
        self.assertEqual(sections["next"], [4])
        pages = pager.get_page_numbers(sections)
        self.assertEqual(pages, [1, 2, 3, 4, 5])

    def test_on_page_5_of_9(self):
        sections = pager.get_sections(page_num=5, page_count=9)
        self.assertEqual(sections["previous"], [4])
        self.assertEqual(sections["first"], [1])
        self.assertEqual(sections["before"], [3, 4])
        self.assertEqual(sections["current"], [5])
        self.assertEqual(sections["after"], [6, 7])
        self.assertEqual(sections["last"], [9])
        self.assertEqual(sections["next"], [6])
        pages = pager.get_page_numbers(sections)
        self.assertEqual(pages, [1, 3, 4, 5, 6, 7, 9])

    def test_on_page_999_of_1000(self):
        sections = pager.get_sections(page_num=999, page_count=1000)
        self.assertEqual(sections["previous"], [998])
        self.assertEqual(sections["first"], [1])
        self.assertEqual(sections["before"], [996, 997, 998])
        self.assertEqual(sections["current"], [999])
        self.assertEqual(sections["after"], [1000])
        self.assertEqual(sections["last"], [1000])
        self.assertEqual(sections["next"], [1000])
        pages = pager.get_page_numbers(sections)
        self.assertEqual(pages, [1, 996, 997, 998, 999, 1000])

    def test_min_page_num(self):
        sections = pager.get_sections(page_num=-1, page_count=2)
        self.assertNotIn("previous", sections)
        self.assertNotIn("first", sections)
        self.assertFalse(sections["before"])
        self.assertEqual(sections["current"], [1])
        self.assertEqual(sections["after"], [2])
        self.assertEqual(sections["last"], [2])
        self.assertEqual(sections["next"], [2])
        pages = pager.get_page_numbers(sections)
        self.assertEqual(pages, [1, 2])

    def test_max_page_num(self):
        sections = pager.get_sections(page_num=3, page_count=2)
        self.assertEqual(sections["previous"], [1])
        self.assertEqual(sections["first"], [1])
        self.assertEqual(sections["before"], [1])
        self.assertEqual(sections["current"], [2])
        self.assertFalse(sections["after"])
        self.assertNotIn("last", sections)
        self.assertNotIn("next", sections)
        pages = pager.get_page_numbers(sections)
        self.assertEqual(pages, [1, 2])

    def test_invalid_page_count(self):
        sections = pager.get_sections(page_num=2, page_count=-1)
        self.assertIsNone(sections)


if __name__ == "__main__":
    unittest.main()



import re
from abc import ABC, abstractmethod
from collections.abc import Iterable
from itertools import chain

from flask_babel import gettext
from whoosh.qparser import MultifieldParser, plugins
from whoosh.query import And, Every, Not, Or, Term
from whoosh.query.ranges import DateRange
from whoosh.sorting import Count, Facets, FieldFacet

from kerko.exceptions import except_raise
from kerko.shortcuts import composer
from kerko.storage import SchemaError


class Searcher:
    """
    Provide a convenient API for searching.

    This adapter wraps a `whoosh.searching.Searcher` instance and implements the
    context management protocol (for `with` statements).
    """

    def __init__(
        self,
        index,
        schema=None,
        field_specs=None,
        facet_specs=None,
    ):
        self.searcher = index.searcher()
        self.schema = schema or composer().schema
        self.field_specs = field_specs or composer().fields
        self.facet_specs = {  # Reorganize by filter key instead of spec key.
            f.filter_key: f
            for f in (facet_specs.values() if facet_specs else composer().facets.values())
        }
        self.search_args = {}  # Arguments to pass to Whoosh's searcher.

    def close(self):
        self.searcher.close()

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self.close()

    @except_raise(KeyError, SchemaError, "Schema changes are required. Please clean index.")
    def search(self, *, limit=None, **kwargs):
        self._prepare_search_args(**kwargs)
        return UnpagedResults(
            self.searcher.search(
                q=self.search_args.pop("query"),
                limit=limit,
                **self.search_args,
            )
        )

    @except_raise(KeyError, SchemaError, "Schema changes are required. Please clean index.")
    def search_page(self, *, page, page_len, **kwargs):
        self._prepare_search_args(**kwargs)
        return PagedResults(
            self.searcher.search_page(
                pagenum=page,
                pagelen=page_len,
                **self.search_args,
            )
        )

    def _prepare_search_args(
        self,
        *,
        keywords=None,
        filters=None,
        require_all=None,
        require_any=None,
        require_date_ranges=None,
        reject_any=None,
        sort_spec=None,
        faceting=False,
    ):
        """
        Prepare search arguments.

        :param bool faceting: Whether to compute groupings along with the search
            results.
        """
        self._prepare_keywords(keywords)
        self._prepare_filters(filters, require_all, require_any, require_date_ranges, reject_any)
        self._prepare_sorting(sort_spec)
        if faceting:
            self._prepare_faceting()

    def _prepare_keywords(self, keywords=None):
        """
        Prepare query parsers.

        :param MultiDict keywords: The search texts keyed by scope key. If falsy,
            the query will match every documents.
        """
        if keywords:
            queries = []
            text_plugins = [
                plugins.PhrasePlugin(),
                plugins.GroupPlugin(),
                plugins.OperatorsPlugin(
                    And=r"(?<=\s)" + re.escape(gettext("AND")) + r"(?=\s)",
                    Or=r"(?<=\s)" + re.escape(gettext("OR")) + r"(?=\s)",
                    Not=r"(^|(?<=(\s|[()])))" + re.escape(gettext("NOT")) + r"(?=\s)",
                    AndNot=None,
                    AndMaybe=None,
                    Require=None,
                ),
                plugins.BoostPlugin(),
            ]
            for key, value in keywords.items(multi=True):
                fields = [
                    field_spec.key
                    for field_spec in self.field_specs.values()
                    if key in field_spec.scopes
                ]
                if not fields:
                    raise KeyError  # No known field for that scope key.
                parser = MultifieldParser(fields, schema=self.schema, plugins=text_plugins)
                queries.append(parser.parse(value))
            self.search_args["query"] = And(queries)
        else:
            self.search_args["query"] = Every()

    def _prepare_filters(
        self,
        filters=None,
        require_all=None,
        require_any=None,
        require_date_ranges=None,
        reject_any=None,
    ):
        """
        Prepare query filtering terms.

        :param MultiDict filters: The facet filters to apply keyed by filter
            key. If falsy, no filters will be applied.

        :param dict require_all: Values that are all required to match, that
            will be combined to the search with the `And` boolean search
            operator. Each key must correspond to a schema field name. The value
            must be a list of one or more accepted values (which will be
            combined together with a nested `Or` boolean search operator).

        :param dict require_any: A dict of inclusion filters, where the key and
            value are, respectively, the schema field name and the list of field
            values to allow in the search results. Any match will allow an item
            in the results, but at least one match is required (values are
            expanded to a flat list of field, value pairs, which are combined
            with the `Or` boolean search operator).

        :param dict reject_any: A dict of exclusion filters, where the key and
            value are, respectively, the schema field name and a list of one or
            more field values to reject. Any match will cause an item to be
            rejected from the results (values are expanded to a flat list of
            field, value pairs, which are combined with the `And` and `Not`
            boolean search operators).

        :param dict require_date_ranges: A dict of date ranges that are all
            required to match. Each key and value are, respectively, the schema
            field name and a tuple with start and end `datetime` objects (the
            range may be open at either end, using `None` instead of a
            `datetime` instance). Items falling outside the date range will be
            rejected from the results.
        """
        terms = []
        if require_all:
            for field_key in require_all.keys() & self.field_specs.keys():
                terms.append(Or([Term(field_key, value) for value in require_all[field_key]]))
        if require_any:
            terms.append(
                Or(
                    list(
                        chain(
                            *[
                                [Term(field_name, value) for value in allow_values]
                                for field_name, allow_values in require_any.items()
                                if field_name in self.field_specs.keys()
                            ]
                        )
                    )
                )
            )
        if require_date_ranges:
            for field_key in require_date_ranges.keys() & self.field_specs.keys():
                terms.append(DateRange(field_key, *require_date_ranges[field_key]))
        if reject_any:
            for field_name, reject_values in reject_any.items():
                if field_name in self.field_specs.keys():
                    terms.extend([Not(Term(field_name, value)) for value in reject_values])
        if filters:
            for filter_key, filter_values in filters.lists():
                spec = self.facet_specs.get(filter_key)
                if spec:
                    for v in filter_values:
                        if v == "":  # If trying to filter with a missing value.
                            # Exclude all results with a value in facet field.
                            terms.append(Not(Every(spec.key)))
                        else:
                            v = spec.codec.transform_for_query(v)  # noqa: PLW2901
                            terms.append(spec.query_class(spec.key, v))
        if terms:
            self.search_args["filter"] = And(terms)

    def _prepare_sorting(self, sort_spec=None):
        if sort_spec and sort_spec.fields:  # Else will sort by relevance score.
            if isinstance(sort_spec.reverse, Iterable):
                # Per-field reverse values require to use `FieldFacet` objects.
                self.search_args["sortedby"] = [
                    FieldFacet(field_key, reverse=reverse)
                    for field_key, reverse in zip(sort_spec.get_field_keys(), sort_spec.reverse)
                ]
            else:
                # With a single reverse value, we may specify the field names directly.
                self.search_args["sortedby"] = sort_spec.get_field_keys()
                self.search_args["reverse"] = sort_spec.reverse

    def _prepare_faceting(self):
        self.search_args["groupedby"] = Facets()
        for facet_spec in self.facet_specs.values():
            self.search_args["groupedby"].add_field(
                facet_spec.key, allow_overlap=facet_spec.allow_overlap
            )
        self.search_args["maptype"] = Count


class Results(ABC):
    """
    Provide a simple unified interface for search results.

    This abstract base class wraps a Whoosh results objects, and concrete
    classes must provide the adaptations for the common interface.
    """

    def __init__(self, results):
        """
        Initialize the adapter.

        :param results: Either a `whoosh.searching.Results`, or a
            `whoosh.searching.ResultsPage` object.
        """
        self._results = results

    def __iter__(self):
        """Yield a `whoosh.searching.Hit` object for each result in ranked order."""
        return self._results.__iter__()

    def __getitem__(self, n):
        return self._results.__getitem__(n)

    def __len__(self):
        return self._results.__len__()

    def is_empty(self):
        return self._results.scored_length() == 0

    def __nonzero__(self):
        return not self.is_empty()

    __bool__ = __nonzero__

    @property
    @abstractmethod
    def page_count(self):
        pass

    @property
    @abstractmethod
    def item_count(self):
        pass

    def items(self, field_specs, facet_specs=None):
        """
        Load search result items, with just the specified fields.

        This may be called only while the searcher object is still alive.

        :param dict field_specs: The specifications of the fields to load into
            each item. Any requested field that is not present in a result is
            silently ignored for that result.

        :param dict facet_specs: The specifications of the facets to load into
            each item. Any requested facet that is not present in a result is
            silently ignored for that result.
        """
        return [self._item(hit, field_specs, facet_specs) for hit in self._results]

    def facets(self, facet_specs, criteria, active_only=False):
        """
        Load facets from search results.

        This may be called only while the searcher object is still alive.

        :param dict facet_specs: The specifications of the facets to load. Any
            requested facet that doesn't have a corresponding grouping in the
            search results is silently ignored.
        """
        return {
            key: facet_specs[key].build(self._groups(key), criteria, active_only)
            for key in facet_specs.keys() & self._facet_names()
        }

    @staticmethod
    def _item(hit, field_specs, facet_specs=None):
        """
        Copy specified fields from a `whoosh.searching.Hit` object into a `dict`.
        """
        facet_specs = facet_specs or {}
        return {
            **{key: field_specs[key].decode(hit[key]) for key in field_specs.keys() & hit.keys()},
            **{key: hit[key] for key in facet_specs.keys() & hit.keys()},
        }

    @abstractmethod
    def _groups(self, name=None):
        """
        Return groups resulting from the 'groupedby' search argument.
        """

    @abstractmethod
    def _facet_names(self):
        pass


class UnpagedResults(Results):
    """Interface results from a `whoosh.searching.Results` object."""

    @property
    def page_count(self):
        return 1

    @property
    def item_count(self):
        return self._results.estimated_length()

    def _groups(self, name=None):
        return self._results.groups(name)

    def _facet_names(self):
        return self._results.facet_names()


class PagedResults(Results):
    """Interface results from a `whoosh.searching.PageResults` object."""

    @property
    def page_count(self):
        return self._results.pagecount

    @property
    def item_count(self):
        return self._results.total

    def _groups(self, name=None):
        return self._results.results.groups(name)

    def _facet_names(self):
        return self._results.results.facet_names()



from flask import url_for

from kerko.shortcuts import composer


def build_sorter(criteria):
    sorter = {}
    active_sort_spec = criteria.get_active_sort_spec()
    if active_sort_spec:
        sorter["active"] = active_sort_spec.label
    sorter["items"] = [
        {
            "label": sort_spec.label,
            "url": url_for(
                ".search",
                **criteria.params(
                    options={
                        "sort": sort_spec.key,
                        "page": None,
                    }
                ),
            ),
        }
        for sort_spec in composer().get_ordered_specs("sorts")
        if sort_spec.is_allowed(criteria)
    ]

    return sorter



"""
Custom filters for Jinja2 templates.
"""

from __future__ import annotations

import re
from urllib.parse import urlparse

from jinja2 import pass_eval_context
from markupsafe import Markup
from w3lib.url import safe_url_string

from kerko import richtext
from kerko.datetime import format_datetime, reformat_date
from kerko.views.item import get_item_title


@pass_eval_context
def urlize_doi(eval_ctx, doi, target=None, rel=None, link=True):
    """Convert the specified DOI to an URL."""
    if not re.match(r"^https?://", doi, flags=re.IGNORECASE):
        url = "https://doi.org/" + str(doi)
    else:
        url = str(doi)
    if link:
        target = f' target="{target}"' if target else ""
        rel = f' rel="{rel}"' if rel else ""
        result = f'<a href="{url}"{target}{rel}>{doi}</a>'
    else:
        result = url
    if eval_ctx.autoescape:
        result = Markup(result)
    return result


@pass_eval_context
def parse_and_urlize_doi(eval_ctx, text, target=None, rel=None):
    """Convert a DOI to an URL, on all lines having the 'DOI:' prefix."""
    target = f' target="{target}"' if target else ""
    rel = f' rel="{rel}"' if rel else ""
    result = re.sub(
        r"^(?P<prefix>\s*DOI:\s*)(?!https?://)(?P<doi>\S+)\s*$",
        r'\g<prefix><a href="https://doi.org/\g<doi>"' + target + rel + r">\g<doi></a>",
        str(text),
        flags=re.IGNORECASE | re.MULTILINE,
    )
    if eval_ctx.autoescape:
        result = Markup(result)
    return result


def register_filters(blueprint):
    """Add custom template filters."""
    blueprint.add_app_template_filter(format_datetime, name="kerko_format_datetime")
    blueprint.add_app_template_filter(reformat_date, name="kerko_reformat_date")
    blueprint.add_app_template_filter(
        lambda text: urlparse(text).hostname, name="kerko_url_hostname"
    )
    blueprint.add_app_template_filter(safe_url_string, name="kerko_url_sanitize")
    blueprint.add_app_template_filter(urlize_doi, name="kerko_urlize_doi")
    blueprint.add_app_template_filter(parse_and_urlize_doi, name="kerko_parse_and_urlize_doi")
    blueprint.add_app_template_filter(richtext.richtext_escape, name="kerko_richtext_escape")


def register_globals(blueprint):
    """Add custom template global functions."""
    blueprint.add_app_template_global(get_item_title, name="kerko_item_title")



"""Download from Zotero the file attachments referenced by the cache."""

import hashlib

from flask import current_app

from kerko.extractors import is_file_attachment
from kerko.searcher import Searcher
from kerko.shortcuts import composer, config
from kerko.storage import get_storage_dir, open_index
from kerko.sync import zotero


def md5_checksum(path):
    # Hash file in 8k blocks to avoid reading it all in memory.
    with path.open("rb") as f:
        md5_hash = hashlib.md5()
        while True:
            data = f.read(8192)
            if not data:
                break
            md5_hash.update(data)
        return md5_hash.hexdigest()


def sync_attachments(full=False):
    """
    Synchronize attachments from Zotero into the attachments directory.

    Files are requested based on item data available in the search index. Thus,
    it always makes sense to synchronize the search index beforehand.

    Return the number of synchronized items.
    """

    def _sync_attachment(attachment, parent=None):
        context = f"(parent item: {parent['id']})" if parent else "(standalone)"
        if not attachment.get("id"):
            current_app.logger.warning(f"An attachment lacks an id {context}. Skipped.")
            return
        if not attachment["data"].get("md5"):
            current_app.logger.warning(f"Attachment {attachment['id']} lacks a checksum {context}.")
        if attachment["id"] in local_files:
            local_files.remove(attachment["id"])
        filepath = attachments_dir / attachment["id"]
        if (
            full
            or not filepath.exists()
            or md5_checksum(filepath) != attachment["data"].get("md5", "")
        ):
            current_app.logger.debug(f"Requesting attachment {attachment['id']} {context}...")
            try:
                # Download attachment.
                with filepath.open("wb") as f:
                    f.write(zotero.retrieve_file(zotero_credentials, attachment["id"]))
            except zotero.zotero_errors.PyZoteroError as e:
                current_app.logger.error(
                    f"Could not download attachment {attachment['id']} {context}: {e}"
                )
        else:
            current_app.logger.debug(f"Keeping attachment {attachment['id']} {context}.")

    current_app.logger.info("Starting attachment files sync...")
    attachments_dir = get_storage_dir("attachments")
    attachments_dir.mkdir(parents=True, exist_ok=True)
    local_files = {p.name for p in attachments_dir.iterdir()}

    # List all items from the search index and request their attachments, if
    # any, from Zotero
    zotero_credentials = zotero.init_zotero()
    index = open_index("index")
    with Searcher(index) as searcher:
        count = 0
        results = searcher.search(limit=None)  # Retrieve all items.
        for item in results.items(
            composer().select_fields(["id", "item_type", "attachments", "data"])
        ):
            if item["item_type"] == "attachment" and is_file_attachment(
                item, config("kerko.zotero.attachment_mime_types")
            ):
                _sync_attachment(item)
                count += 1
            else:
                # Child attachments, unlike standalone attachments (above), do not
                # need their linkMode or MIME type to be validated, because that has
                # already been done by an extractor when writing the search index.
                for attachment in item.get("attachments", []):
                    _sync_attachment(attachment, parent=item)
                    count += 1

    # Delete remaining local files that were not referenced by any item.
    for name in local_files:
        current_app.logger.debug(f"Deleting attachment {name}, unused.")
        (attachments_dir / name).unlink()

    current_app.logger.info(f"Attachment files sync successful ({count} file(s) processed).")
    return count


def delete_attachments():
    attachments_dir = get_storage_dir("attachments")
    if attachments_dir.is_dir():
        for filepath in attachments_dir.iterdir():
            filepath.unlink()
        attachments_dir.rmdir()



from kerko.views import routes

# List of URL rules, defined as dicts of arguments to pass to `Blueprint.add_url_rule()`.
urls = [
    {
        "rule": "/",
        "view_func": routes.search,
        "methods": ["GET", "POST"],
    },
    {
        "rule": "/atom.xml",
        "view_func": routes.atom_feed,
        "methods": ["GET"],
    },
    {
        "rule": "/<path:item_id>",
        "view_func": routes.item_view,
    },
    {
        "rule": "/<path:item_id>/download/<string:attachment_id>/",
        "view_func": routes.child_attachment_download,
    },
    {
        "rule": "/<path:item_id>/download/<string:attachment_id>/<string:attachment_filename>",
        "view_func": routes.child_attachment_download,
    },
    {
        "rule": "/download/<string:item_id>/",
        "view_func": routes.standalone_attachment_download,
    },
    {
        "rule": "/download/<string:item_id>/<string:attachment_filename>",
        "view_func": routes.standalone_attachment_download,
    },
    {
        "rule": "/<path:item_id>/export/<string:bib_format_key>",
        "view_func": routes.item_bib_download,
    },
    {
        "rule": "/export/<string:bib_format_key>/",
        "view_func": routes.search_bib_download,
    },
    {
        "rule": "/sitemap.xml",
        "view_func": routes.sitemap_index,
    },
    {
        "rule": "/sitemap<int(min=1):page_num>.xml",
        "view_func": routes.sitemap,
    },
    {
        "rule": "/api/last-sync",
        "view_func": routes.last_updated_on,
    },
]



"""
Unit tests for the html module.
"""

import unittest

from markupsafe import Markup

from kerko import richtext


class RichTextEscapeTestCase(unittest.TestCase):
    """Test the ``richtext_escape()`` function."""

    def test_safe_markup(self):
        s = Markup("Hello, <em>world</em>! '\"<>#&amp;<p></p><span></span><em></em><br><br/>")
        e = richtext.richtext_escape(s)
        self.assertEqual(e, s)
        self.assertIsInstance(e, Markup)

    def test_escaped_text(self):
        e = richtext.richtext_escape("Hello, Zotero & Kerko!")
        self.assertEqual(
            e,
            "Hello, Zotero &amp; Kerko!",
        )
        self.assertIsInstance(e, Markup)

        e = richtext.richtext_escape("'\"<>#&amp;&#39;")
        self.assertEqual(
            e,
            "&#39;&#34;&lt;&gt;#&amp;amp;&amp;#39;",
        )
        self.assertIsInstance(e, Markup)

    def test_escaped_tags(self):
        e = richtext.richtext_escape("Hello, <em>world</em>!")
        self.assertEqual(e, "Hello, &lt;em&gt;world&lt;/em&gt;!")
        self.assertIsInstance(e, Markup)

        e = richtext.richtext_escape("<div><strong><br /></strong></div>")
        self.assertEqual(
            e,
            "&lt;div&gt;&lt;strong&gt;&lt;br /&gt;&lt;/strong&gt;&lt;/div&gt;",
        )
        self.assertIsInstance(e, Markup)

        e = richtext.richtext_escape("<script>alert(1)</script>")
        self.assertEqual(
            e,
            "&lt;script&gt;alert(1)&lt;/script&gt;",
        )
        self.assertIsInstance(e, Markup)

        e = richtext.richtext_escape("<span>Hello, world!</span>")
        self.assertEqual(
            e,
            "&lt;span&gt;Hello, world!&lt;/span&gt;",
        )
        self.assertIsInstance(e, Markup)

    def test_unescaped_tags(self):
        s = "<b>Hello</b>, <i>world</i>!"
        self.assertEqual(richtext.richtext_escape(s), Markup(s))

        s = "<sub>sub 1</sub><sub>sub 2</sub><sup>sup 1</sup><sub>sub 3</sub>"
        self.assertEqual(richtext.richtext_escape(s), Markup(s))

        s = '<span class="nocase">Hello, world!</span>'
        self.assertEqual(richtext.richtext_escape(s), Markup(s))

        s = (
            '<span style="font-variant:small-caps;">Small Caps 1</span>'
            '<span style="font-variant:small-caps;">Small Caps 2</span>'
        )
        self.assertEqual(richtext.richtext_escape(s), Markup(s))

    def test_escaped_and_unescaped_combination(self):
        self.assertEqual(
            richtext.richtext_escape(
                "<span><b>Span escaped, nested b not escaped</b></span>"
                '<span style="font-variant:small-caps;">Span <i>not</i> escaped</span>'
            ),
            Markup(
                "&lt;span&gt;<b>Span escaped, nested b not escaped</b>&lt;/span&gt;"
                '<span style="font-variant:small-caps;">Span <i>not</i> escaped</span>'
            ),
        )


class RichTextStriptagsTestCase(unittest.TestCase):
    """Test the ``richtext_striptags()`` function."""

    def test_richtext_tags(self):
        s = "<b>Hello</b>, <i>world</i>!"
        self.assertEqual(richtext.richtext_striptags(s), "Hello, world!")

        s = "<sub>sub 1</sub><sub>sub 2</sub><sup>sup 1</sup><sub>sub 3</sub>"
        self.assertEqual(richtext.richtext_striptags(s), "sub 1sub 2sup 1sub 3")

        s = '<span class="nocase">Hello, world!</span>'
        self.assertEqual(richtext.richtext_striptags(s), "Hello, world!")

        s = (
            '<span style="font-variant:small-caps;">Small Caps 1</span>'
            '<span style="font-variant:small-caps;">Small Caps 2</span>'
        )
        self.assertEqual(richtext.richtext_striptags(s), "Small Caps 1Small Caps 2")

    def test_html_tags(self):
        s = (
            '<span style="font-variant:small-caps;"><em>emphasis</em> '
            "and <strong>strong</strong></span>"
        )
        self.assertEqual(
            richtext.richtext_striptags(s), "<em>emphasis</em> and <strong>strong</strong>"
        )

    def test_text(self):
        s = "&amp; &amp;amp; &lt;b&gt;b&lt;/b&gt;"
        self.assertEqual(richtext.richtext_striptags(s), s)

        s = '"2 > 1" & "1 < 2"'
        self.assertEqual(richtext.richtext_striptags(s), s)

        s = "<this is no HTML>"
        self.assertEqual(richtext.richtext_striptags(s), s)



"""Update the search index from the local cache."""

import whoosh
from flask import current_app
from whoosh.query import Term

from kerko.extractors import ItemTitleExtractor
from kerko.shortcuts import composer, config
from kerko.storage import SchemaError, SearchIndexError, load_object, open_index, save_object
from kerko.tags import TagGate


def sync_index(full=False):  # noqa: ARG001
    """
    Build the search index from the local cache.

    Return the number of synchronized items.
    """
    current_app.logger.info("Starting index sync...")
    library_context = load_object("cache", "library")

    cache = open_index("cache")
    cache_version = load_object("cache", "version", default=0)
    if not cache_version:
        msg = "The cache is empty and needs to be synchronized first."
        raise SearchIndexError(msg)

    # FIXME: The following does not detect when just the collections have changed in the cache
    #        (with no item changes). Should check the collections_version!
    #        https://pyzotero.readthedocs.io/en/latest/#zotero.Zotero.collection_versions
    # if not full and load_object('index', 'version', default=0) == cache_version:
    #     current_app.logger.warning(
    #         f"The index is already up-to-date with cache version {cache_version}, nothing to do."
    #     )
    #     return 0

    def yield_items(parent_key):
        with cache.searcher() as searcher:
            results = searcher.search(Term("parentItem", parent_key), limit=None)
            for hit in results:
                yield hit.fields()

    def yield_top_level_items():
        return yield_items("")

    def yield_children(parent):
        return yield_items(parent["key"])

    count = 0
    index = open_index("index", schema=composer().schema, auto_create=True, write=True)
    writer = index.writer(
        limitmb=config("kerko.performance.whoosh_index_memory_limit"),
        procs=config("kerko.performance.whoosh_index_processors"),
    )
    try:
        writer.mergetype = whoosh.writing.CLEAR
        gate = TagGate(
            config("kerko.zotero.item_include_re"),
            config("kerko.zotero.item_exclude_re"),
        )
        for item in yield_top_level_items():
            count += 1
            if gate.check(item["data"]):
                item["children"] = list(yield_children(item))  # Extend the base Zotero item dict.
                document = {}
                for spec in list(composer().fields.values()) + list(composer().facets.values()):
                    spec.extract_to_document(document, item, library_context)
                writer.update_document(**document)
                current_app.logger.debug(
                    f"Item {count} updated ({item['key']}, {item.get('itemType')}): "
                    f"{ItemTitleExtractor().extract(item, library_context, None)}"
                )
            else:
                current_app.logger.debug(f"Item {count} excluded ({item['key']})")
    except (whoosh.fields.FieldConfigurationError, whoosh.fields.UnknownFieldError) as e:
        writer.cancel()
        current_app.logger.error(e)
        msg = "Schema changes are required. Please clean index."
        raise SchemaError(msg) from e
    except Exception:
        writer.cancel()
        raise
    else:
        writer.commit()
        # Save the cache's last_modified timestamp. Later, we cannot access the
        # cache directly to show the user when the data was last synchronized
        # from Zotero, because there is no guarantee that what the user's sees
        # (i.e., the current content of the index) is still in sync with the
        # cache (the cache might have been cleaned, or it might have been just
        # updated, with an index update still pending).
        save_object("index", "last_update_from_zotero", cache.last_modified())
        save_object("index", "version", cache_version)
        current_app.logger.info(
            f"Index sync successful, now at version {cache_version} "
            f"({count} top level item(s) processed)."
        )
    return count



import pprint
from datetime import datetime
from typing import Any

import click
import tomli_w
import wrapt
from flask import current_app
from flask.cli import with_appcontext

from kerko.config_helpers import is_toml_serializable
from kerko.storage import SchemaError, SearchIndexError, delete_storage, get_doc_count
from kerko.sync import zotero
from kerko.sync.attachments import delete_attachments, sync_attachments
from kerko.sync.cache import sync_cache
from kerko.sync.index import sync_index


@wrapt.decorator
def execution_time_logger(wrapped, _instance, args, kwargs):
    start_time = datetime.now()
    return_value = wrapped(*args, **kwargs)
    current_app.logger.info(_format_elapsed_time(start_time))
    return return_value


@click.group()
def cli():
    """Run a Kerko subcommand."""


@cli.command()
@click.argument(
    "target",
    default="everything",
    type=click.Choice(["cache", "index", "attachments", "everything"], case_sensitive=False),
)
@click.option(
    "--full",
    default=False,
    is_flag=True,
    flag_value=True,
    help=(
        "When possible, the synchronization process performs an incremental "
        "update of just the new or changed items since the last "
        "synchronization. This option forces a full update."
    ),
)
@with_appcontext
@execution_time_logger
def sync(target, full=False):
    """
    Synchronize the cache, the search index, and/or the file attachments.

    By default, everything is synchronized.
    """
    try:
        if target in ["everything", "cache"]:
            sync_cache(full)
        if target in ["everything", "index"]:
            sync_index(full)
        if target in ["everything", "attachments"]:
            sync_attachments(full)
    except SearchIndexError as e:
        current_app.logger.error(e)
        raise click.Abort
    except SchemaError as e:
        current_app.logger.error(e)
        raise click.Abort


@cli.command()
@click.argument(
    "target",
    type=click.Choice(["cache", "index", "attachments", "everything"], case_sensitive=False),
)
@with_appcontext
def clean(target):
    """
    Delete the specified data.

    Use the argument to select which data to delete, either the cache, the
    search index, the attachments, or all of those (everything).
    """
    if target in ["everything", "cache"]:
        delete_storage("cache")
    if target in ["everything", "index"]:
        delete_storage("index")
    if target in ["everything", "attachments"]:
        delete_attachments()


@cli.command()
@click.argument(
    "target",
    type=click.Choice(["cache", "index"], case_sensitive=False),
)
@with_appcontext
def count(target):
    """
    Show the number of records available in the cache or in the search index.

    The cache and the index are structured very differently and their respective
    numbers should not be expected to match. The number of records in the index
    may not even match the number of results obtained in Kerko's search
    interface, because the search can do some internal filtering.

    The cache is a flat database where items of any type or hierarchical level
    are counted as separate records.

    The index is a flat, denormalized database where each item is grouped with
    its children in a single record. The count may include items that are not
    usually displayed in search results, e.g., standalone notes or attachments.

    WARNING: This command is provided for development purposes only and may be
    modified or removed from the module at any time.
    """
    try:
        click.echo(get_doc_count(target))
    except SearchIndexError as e:
        current_app.logger.error(e)
        raise click.Abort


@cli.command()
@click.option(
    "--show-secrets",
    default=False,
    is_flag=True,
    flag_value=True,
    help="Secrets are hidden from the output by default. This option causes them to be revealed.",
)
@with_appcontext
def config(show_secrets=False):
    """
    Show the configuration.

    Note that parameters that internally have 'None' values will be omitted
    because such values cannot be represented in TOML files.
    """

    def hide_secrets(d: dict):
        for k in d.keys():
            if k in ["SECRET_KEY", "ZOTERO_API_KEY"] or k.find("PASSWORD") >= 0:
                d[k] = "*****"

    def copy_serializable(obj: Any) -> Any:
        """
        Copy the object, with some twists.

        - Filter values that cannot be serialized as TOML.
        - Sort dicts by key.
        """
        if isinstance(obj, dict):
            new_dict = {}
            for k, v in sorted(obj.items()):
                new_v = copy_serializable(v)
                if new_v is not None:
                    new_dict[k] = new_v
            return new_dict
        elif isinstance(obj, list):
            new_list = []
            for v in obj:
                new_v = copy_serializable(v)
                if new_v is not None:
                    new_list.append(new_v)
            return new_list
        elif is_toml_serializable(obj):
            return obj

    serializable_config = copy_serializable(current_app.config)
    if not show_secrets:
        hide_secrets(serializable_config)
    click.echo(tomli_w.dumps(serializable_config))


@cli.command()
@click.argument("item_key")
@with_appcontext
def zotero_item(item_key):
    """
    Retrieve an item from the library, using the Zotero API.

    WARNING: This command is provided for development purposes only and may be
    modified or removed from the module at any time.
    """
    credentials = zotero.init_zotero()
    click.echo(pprint.pformat(zotero.load_item(credentials, item_key)))


@cli.command()
@with_appcontext
def zotero_item_types():
    """
    List all item types, using the Zotero API.

    WARNING: This command is provided for development purposes only and may be
    modified or removed from the module at any time.
    """
    credentials = zotero.init_zotero()
    click.echo(pprint.pformat(zotero.load_item_types(credentials)))


@cli.command()
@with_appcontext
def zotero_item_fields():
    """
    List all fields, using the Zotero API.

    WARNING: This command is provided for development purposes only and may be
    modified or removed from the module at any time.
    """
    credentials = zotero.init_zotero()
    click.echo(pprint.pformat(zotero.load_item_fields(credentials)))


@cli.command()
@click.argument("item_type")
@with_appcontext
def zotero_item_type_fields(item_type):
    """
    List the available fields for a given item type, using the Zotero API.

    WARNING: This command is provided for development purposes only and may be
    modified or removed from the module at any time.
    """
    credentials = zotero.init_zotero()
    click.echo(pprint.pformat(zotero.load_item_type_fields(credentials, item_type)))


@cli.command()
@click.argument("item_type")
@with_appcontext
def zotero_item_type_creator_types(item_type):
    """
    List the available creator types for a given item type, using the Zotero API.

    WARNING: This command is provided for development purposes only and may be
    modified or removed from the module at any time.
    """
    credentials = zotero.init_zotero()
    click.echo(pprint.pformat(zotero.load_item_type_creator_types(credentials, item_type)))


@cli.command()
@with_appcontext
def zotero_top_level_collections():
    """
    List top-level collections of the library, using the Zotero API.

    WARNING: This command is provided for development purposes only and may be
    modified or removed from the module at any time.
    """
    credentials = zotero.init_zotero()
    collections = zotero.Collections(credentials, top_level=True)
    for c in collections:
        click.echo(f"{c.get('key')} {c.get('data', {}).get('name', '')}")


def _format_elapsed_time(start_time):
    elapsed_time = int(round((datetime.now() - start_time).total_seconds()))
    elapsed_min, elapsed_sec = elapsed_time // 60, elapsed_time % 60
    s = "Execution time:"
    if elapsed_min > 0:
        s += (" {n} minutes" if elapsed_min > 1 else " {n} minute").format(n=elapsed_min)
        s += (" {n:02} seconds" if elapsed_sec > 1 else " {n:02d} second").format(n=elapsed_sec)
    else:
        s += (" {n} seconds" if elapsed_sec > 1 else " {n} second").format(n=elapsed_sec)
    return s



"""Synchronize the Zotero library into a local cache."""

from flask import current_app
from whoosh.fields import ID, NUMERIC, STORED, FieldConfigurationError, Schema, UnknownFieldError
from whoosh.qparser import QueryParser

from kerko.shortcuts import composer, config
from kerko.storage import SchemaError, load_object, open_index, save_object
from kerko.sync import zotero


def get_formats():
    return {
        spec.extractor.format
        for spec in list(composer().fields.values()) + list(composer().facets.values())
        if spec.extractor.format != "data"
    }


def get_cache_schema():
    # CAUTION: When changing this schema, consider adapting any code that depend
    # on the changes to raise `SchemaError` if the the schema is incorrect.
    schema = Schema(
        key=ID(unique=True, stored=True),  # Copied from Zotero.
        version=NUMERIC(stored=True),  # Copied from Zotero.
        parentItem=ID(stored=True),  # Kerko addition.
        itemType=ID(stored=True),  # Kerko addition.
        library=STORED,  # Copied from Zotero.
        links=STORED,  # Copied from Zotero.
        meta=STORED,  # Copied from Zotero.
        data=STORED,  # Copied from Zotero.
        fulltext=STORED,  # Kerko addition.
    )
    for format_ in get_formats():
        schema.add(format_, STORED)
    return schema


def sync_cache(full=False):
    """
    Build a cache of items retrieved from Zotero.

    Return the number of synchronized items.
    """
    current_app.logger.info("Starting cache sync...")
    count = 0
    zotero_credentials = zotero.init_zotero()
    library_context = zotero.request_library_context(
        zotero_credentials
    )  # TODO: Load pickle & sync collections incrementally
    since = load_object("cache", "version", default=0) if not full else 0
    version = zotero.last_modified_version(zotero_credentials)

    cache = open_index("cache", schema=get_cache_schema, auto_create=True, write=True)
    writer = cache.writer(
        limitmb=config("kerko.performance.whoosh_index_memory_limit"),
        procs=config("kerko.performance.whoosh_index_processors"),
    )
    try:
        if config("kerko.search.fulltext"):
            fulltext_items = zotero.load_new_fulltext(zotero_credentials, since)
        else:
            fulltext_items = {}

        formats = get_formats()
        for item in zotero.Items(zotero_credentials, since=since, formats=list(formats) + ["data"]):
            count += 1

            document = {
                "key": item.get("key"),
                "version": item.get("version"),
                "parentItem": item.get("data", {}).get("parentItem", ""),
                "itemType": item.get("data", {}).get("itemType", ""),
                "library": item.get("library", {}),
                "links": item.get("links", {}),
                "meta": item.get("meta", {}),
                "data": item.get("data", {}),
            }
            for format_ in formats:
                if format_ in item:
                    document[format_] = item[format_]
            if item.get("key") in fulltext_items:
                del fulltext_items[item.get("key")]  # Mark this fulltext as updated.
                fulltext = zotero.load_item_fulltext(zotero_credentials, item.get("key"))
                if fulltext:
                    document["fulltext"] = fulltext

            writer.update_document(**document)
            current_app.logger.debug(
                f"Item {count} updated ({item.get('key')}, version {item.get('version')})"
            )

        # Retrieve the updated fulltext of items that were otherwise unchanged.
        for item_key in fulltext_items.keys():
            with cache.searcher() as searcher:
                results = searcher.search(
                    QueryParser("key", schema=cache.schema, plugins=[]).parse(item_key), limit=1
                )
                if results:
                    count += 1
                    document = results[0].fields()
                    fulltext = zotero.load_item_fulltext(zotero_credentials, item_key)
                    if fulltext:
                        document["fulltext"] = fulltext
                    elif "fulltext" in document:
                        del document["fulltext"]
                    writer.update_document(**document)
                    current_app.logger.debug(
                        f"Item {count} text content updated "
                        f"({item_key}, version {document['version']})"
                    )

        # Check for items to remove.
        if since > 0:
            for deleted in zotero.load_deleted_or_trashed_items(zotero_credentials, since):
                count += 1
                writer.delete_by_term("key", deleted)
                current_app.logger.debug(f"Item {count} removed ({deleted})")
    except (FieldConfigurationError, UnknownFieldError) as e:
        writer.cancel()
        current_app.logger.error(e)
        msg = "Schema changes are required. Please clean cache."
        raise SchemaError(msg) from e
    except Exception:
        writer.cancel()
        raise
    else:
        writer.commit()
        save_object("cache", "version", version)
        save_object("cache", "library", library_context)
        current_app.logger.info(
            f"Cache sync successful, now at version {version} ({count} item(s) processed)."
        )
    return count



import math
import time
from collections import deque
from datetime import datetime, timedelta

from flask import (
    abort,
    current_app,
    flash,
    make_response,
    redirect,
    render_template,
    request,
    send_from_directory,
    url_for,
)
from flask_babel import get_locale, gettext
from markupsafe import Markup

from kerko.criteria import create_feed_criteria, create_search_criteria
from kerko.exceptions import except_abort
from kerko.forms import SearchForm
from kerko.searcher import Searcher
from kerko.shortcuts import composer, config
from kerko.specs import SortSpec
from kerko.storage import (
    SchemaError,
    SearchIndexError,
    get_doc_count,
    get_storage_dir,
    load_object,
    open_index,
)
from kerko.views import pager
from kerko.views.item import build_item_context, creators, inject_item_data
from kerko.views.search import search_list, search_single

# Google allows up to 50k URLs per sitemap, but we prefer shorter pages to
# respond more quickly. The sitemap index ties all those pages together.
SITEMAP_URL_MAX_COUNT = 1000


@except_abort(SchemaError, 500)
@except_abort(SearchIndexError, 503)
def search():
    """View the results of a search."""
    criteria = create_search_criteria(request.args)
    form = SearchForm(csrf_enabled=False)
    if form.validate_on_submit():
        scope = composer().scopes.get(form.scope.data)
        if scope:
            # Add the newly submitted keywords, and redirect to search with them.
            url = url_for(
                ".search",
                **criteria.params(
                    keywords=scope.add_keywords(form.keywords.data, criteria.keywords),
                    options={"page": None},
                ),
            )
            return redirect(url, 302)

    if criteria.options.get("page-len", config("kerko.pagination.page_len")) == 1:
        return search_single(criteria, form)
    return search_list(criteria, form)


@except_abort(SchemaError, 500)
@except_abort(SearchIndexError, 503)
def atom_feed():
    """Build a feed based on the search criteria."""

    if "atom" not in config("kerko.feeds.formats"):
        return abort(404)

    context = {}
    criteria = create_feed_criteria(request.args)
    index = open_index("index")
    with Searcher(index) as searcher:
        extra_args = {}

        sort_field = composer().fields.get("sort_date_added")
        if sort_field:
            extra_args["sort_spec"] = SortSpec(
                key="date_added_desc",
                label="",
                fields=[sort_field],
                reverse=True,
            )
        else:
            current_app.logger.warning(
                "Feed cannot be sorted because the 'sort_date_added' was "
                "removed from the configuration."
            )

        if config("kerko.feeds.max_days"):
            if "filter_date" in composer().fields:
                today = datetime.today()
                start = datetime(today.year, today.month, today.day) - timedelta(
                    config("kerko.feeds.max_days")
                )
                current_app.logger.debug(
                    f"Show items dated {start} and newer"
                    f" (kerko.feeds.max_days == {config('kerko.feeds.max_days')})."
                )
                extra_args["require_date_ranges"] = {"filter_date": (start, None)}
            else:
                current_app.logger.warning(
                    "'kerko.feeds.max_days' is set but has no effect because the "
                    "'filter_date' field was removed from the configuration."
                )

        results = searcher.search_page(
            page=criteria.options.get("page", 1),
            page_len=criteria.options.get("page-len", config("kerko.pagination.page_len")),
            keywords=criteria.keywords,
            filters=criteria.filters,
            require_any=config("kerko.feeds.require_any"),  # Apply custom filtering, if not None.
            reject_any={"item_type": ["note", "attachment"], **config("kerko.feeds.reject_any")},
            faceting=False,
            **extra_args,
        )
        if results.is_empty():
            items = []
        else:
            items = results.items(composer().select_fields(config("kerko.feeds.fields")))
            for item in items:
                creators.inject_creator_display_names(item)
        context["items"] = items
        context["total_count"] = results.item_count
        criteria.fit_page(results.page_count or 1)
        pager_sections = pager.get_sections(criteria.options["page"], results.page_count or 1)

    last_sync = load_object("index", "last_update_from_zotero")
    if last_sync:
        context["last_sync"] = datetime.fromtimestamp(
            last_sync, tz=datetime.now().astimezone().tzinfo
        ).isoformat()
    else:
        context["last_sync"] = datetime.now().isoformat()

    if criteria.is_searching():
        context["feed_title"] = gettext("Custom feed")
    else:
        context["feed_title"] = gettext("Main feed")

    response = make_response(
        render_template(
            config("kerko.templates.atom_feed"),
            pager=pager.build_pager(
                pager_sections, criteria, endpoint="kerko.atom_feed", _external=True
            ),
            page_len=config("kerko.pagination.page_len"),
            feed_url=url_for(
                ".atom_feed",
                _external=True,
                **criteria.params(
                    options={
                        "page": criteria.options["page"] if criteria.options["page"] > 1 else None
                    }
                ),
            ),
            html_url=url_for(
                ".search",
                _external=True,
                **criteria.params(options={"page": None}),
            ),
            is_searching=criteria.is_searching(),
            locale=get_locale(),
            **context,
        )
    )
    response.headers["Content-Type"] = "application/atom+xml; charset=utf-8"
    return response


@except_abort(SchemaError, 500)
@except_abort(SearchIndexError, 503)
def item_view(item_id):
    """View a full bibliographic record."""
    start_time = time.process_time()

    index = open_index("index")
    with Searcher(index) as searcher:
        # Try matching the item by id, with fallback to alternate id.
        try_id_fields = deque(["id", "alternate_id"])
        fellback = False
        while try_id_fields:
            results = searcher.search(
                require_all={try_id_fields.popleft(): [item_id]},
                reject_any={"item_type": ["note", "attachment"]},
                limit=1,
                faceting=False,
            )
            if results.is_empty():
                fellback = True  # Not found on first attempt.
                continue
            break  # Found item, or no more id fields to try.
        if results.is_empty():
            return abort(404)
        item = results.items(composer().fields, composer().facets)[0]
        if fellback:
            return redirect(url_for(".item_view", item_id=item["id"]), 301)
    inject_item_data(item)
    return render_template(
        config("kerko.templates.item"),
        **build_item_context(item),
        time=time.process_time() - start_time,
        locale=get_locale(),
    )


@except_abort(SchemaError, 500)
@except_abort(SearchIndexError, 503)
def page(item_id, title):
    """Render a simple page, using a Zotero note as content."""
    index = open_index("index")
    with Searcher(index) as searcher:
        results = searcher.search(
            require_all={
                "id": [item_id],
                "item_type": ["note"],
            },
            limit=1,
        )
        if results.is_empty():
            return abort(404)
        item = results.items(composer().select_fields(["id", "data"]))[0]
        note = item["data"].get("note", "")
    return render_template(
        config("kerko.templates.page"),
        title=title,
        content=Markup(note),
        locale=get_locale(),
    )


@except_abort(SchemaError, 500)
@except_abort(SearchIndexError, 503)
def child_attachment_download(item_id, attachment_id, attachment_filename=None):
    """
    Download a child attachment.

    If the URL does not specify the attachment's filename or provides the wrong
    filename, a redirect is performed to a corrected URL so that the client gets
    a proper filename.
    """
    index = open_index("index")
    with Searcher(index) as searcher:
        # Try matching the item by id, with fallback to alternate id.
        try_id_fields = deque(["id", "alternate_id"])
        fellback = False
        while try_id_fields:
            results = searcher.search(
                require_all={try_id_fields.popleft(): [item_id]},
                reject_any={"item_type": ["note", "attachment"]},
                limit=1,
                faceting=False,
            )
            if results.is_empty():
                fellback = True  # Not found on first attempt.
                continue
            break  # Found item, or no more id fields to try.
        if results.is_empty():
            return abort(404)
        item = results.items(composer().select_fields(["id", "attachments"]))[0]
    matching_attachments = [a for a in item.get("attachments", []) if a.get("id") == attachment_id]
    if not matching_attachments or len(matching_attachments) > 1:
        flash(
            gettext(
                "The document you have requested has been removed. "
                "Please check below for the latest documents available."
            ),
            "warning",
        )
        return redirect(url_for(".item_view", item_id=item["id"]), 301)
    attachment = matching_attachments[0]

    filepath = get_storage_dir("attachments") / attachment_id
    if not filepath.exists():
        return abort(404)

    filename = attachment["data"].get("filename", item["id"])
    if fellback or attachment_filename != filename:
        return redirect(
            url_for(
                ".child_attachment_download",
                item_id=item["id"],
                attachment_id=attachment_id,
                attachment_filename=filename,
            ),
            301,
        )

    return send_from_directory(
        get_storage_dir("attachments"),
        attachment_id,
        download_name=filename,
        mimetype=attachment["data"].get("contentType", "octet-stream"),
    )


@except_abort(SchemaError, 500)
@except_abort(SearchIndexError, 503)
def standalone_attachment_download(item_id, attachment_filename=None):
    """
    Download a standalone attachment.

    If the URL does not specify the attachment's filename or provides the wrong
    filename, a redirect is performed to a corrected URL so that the client gets
    a proper filename.
    """
    index = open_index("index")
    with Searcher(index) as searcher:
        # Try matching the item by id, with fallback to alternate id.
        try_id_fields = deque(["id", "alternate_id"])
        fellback = False
        while try_id_fields:
            results = searcher.search(
                require_all={
                    try_id_fields.popleft(): [item_id],
                    "item_type": ["attachment"],
                },
                limit=1,
                faceting=False,
            )
            if results.is_empty():
                fellback = True  # Not found on first attempt.
                continue
            break  # Found item, or no more id fields to try.
        if results.is_empty():
            return abort(404)
        item = results.items(composer().select_fields(["id", "data"]))[0]

    filepath = get_storage_dir("attachments") / item["id"]
    if not filepath.exists():
        return abort(404)

    filename = item["data"].get("filename", item["id"])
    if fellback or attachment_filename != filename:
        return redirect(
            url_for(
                ".standalone_attachment_download",
                item_id=item["id"],
                attachment_filename=filename,
            ),
            301,
        )

    return send_from_directory(
        get_storage_dir("attachments"),
        item["id"],
        download_name=filename,
        mimetype=item["data"].get("contentType", "octet-stream"),
    )


@except_abort(SchemaError, 500)
@except_abort(SearchIndexError, 503)
def item_bib_download(item_id, bib_format_key):
    """Download a record."""
    if not config("kerko.features.download_item"):
        return abort(404)

    bib_format = composer().bib_formats.get(bib_format_key)
    if not bib_format:
        return abort(404)

    index = open_index("index")
    with Searcher(index) as searcher:
        # Try matching the item by id, with fallback to alternate id.
        try_id_fields = deque(["id", "alternate_id"])
        fellback = False
        while try_id_fields:
            results = searcher.search(
                require_all={try_id_fields.popleft(): [item_id]},
                reject_any={"item_type": ["note", "attachment"]},
                limit=1,
                faceting=False,
            )
            if results.is_empty():
                fellback = True  # Not found on first attempt.
                continue
            break  # Found item, or no more id fields to try.
        if results.is_empty():
            return abort(404)
        item = results.items(composer().select_fields(["id", bib_format.field.key]))[0]

    content = item.get(bib_format.field.key)
    if not content:
        return abort(404)

    if fellback:
        return redirect(
            url_for(".item_bib_download", item_id=item["id"], bib_format_key=bib_format_key), 301
        )

    response = make_response(content)
    response.headers["Content-Disposition"] = (
        f"attachment; filename={item['id']}.{bib_format.extension}"
    )
    response.headers["Content-Type"] = f"{bib_format.mime_type}; charset=utf-8"
    return response


@except_abort(SchemaError, 500)
@except_abort(SearchIndexError, 503)
def search_bib_download(bib_format_key):
    """Download all records resulting from a search."""
    bib_format = composer().bib_formats.get(bib_format_key)
    if not bib_format:
        return abort(404)

    criteria = create_search_criteria(request.args)
    index = open_index("index")
    with Searcher(index) as searcher:
        results = searcher.search(
            limit=None,
            keywords=criteria.keywords,
            filters=criteria.filters,
            reject_any={"item_type": ["note", "attachment"]},
            sort_spec=criteria.get_active_sort_spec(),
            faceting=False,
        )
        if results.is_empty():
            return abort(404)
        citations = [
            item.get(bib_format.field.key, "")
            for item in results.items(field_specs={bib_format.field.key: bib_format.field})
        ]
    response = make_response(
        bib_format.group_format.format(bib_format.group_item_delimiter.join(citations))
    )
    response.headers["Content-Disposition"] = (
        f"attachment; filename=bibliography.{bib_format.extension}"
    )
    response.headers["Content-Type"] = f"{bib_format.mime_type}; charset=utf-8"
    return response


def get_sitemap_page_count():
    count = get_doc_count("index")
    if count:
        count = math.ceil(count / SITEMAP_URL_MAX_COUNT)
        # Note: count may include items that will be excluded from the
        # sitemap, but this is the efficient way to get a good enough count.
    return count


@except_abort(SearchIndexError, 503)
def sitemap_index():
    """Generate a sitemap index."""
    response = make_response(
        render_template(
            "kerko/sitemap_index.xml.jinja2",
            sitemap_count=get_sitemap_page_count(),
        )
    )
    response.headers["Content-Type"] = "application/xml; charset=utf-8"
    return response


@except_abort(SchemaError, 500)
@except_abort(SearchIndexError, 503)
def sitemap(page_num):
    """Generate a sitemap."""
    if page_num > get_sitemap_page_count():
        return abort(404)

    index = open_index("index")
    with Searcher(index) as searcher:
        sort_field = composer().fields.get("sort_date_modified")
        if sort_field:
            sort_spec = SortSpec(
                key="date_added_desc",
                label="",
                fields=[sort_field],
                reverse=True,
            )
        else:
            sort_spec = None
        results = searcher.search_page(
            page=page_num,
            page_len=SITEMAP_URL_MAX_COUNT,
            reject_any={"item_type": ["note", "attachment"]},
            sort_spec=sort_spec,
            faceting=False,
        )
        if results.is_empty():
            return abort(404)
        items = results.items(composer().select_fields(["id", "date_modified"]))
    response = make_response(render_template("kerko/sitemap.xml.jinja2", items=items))
    response.headers["Content-Type"] = "application/xml; charset=utf-8"
    return response


def last_updated_on():
    last_sync = load_object("index", "last_update_from_zotero")
    if last_sync:
        return {
            "when": datetime.fromtimestamp(
                last_sync, tz=datetime.now().astimezone().tzinfo
            ).isoformat(),
            "hours_ago": round((time.time() - last_sync) / 3600, 3),
        }
    return {}



"""
Unit tests for configuration helpers.
"""

import unittest

from flask import Config

from kerko import DEFAULTS as KERKO_DEFAULTS
from kerko.config_helpers import (
    ConfigModel,
    KerkoModel,
    config_get,
    config_set,
    config_update,
    parse_config,
)


class ConfigUpdateTestCase(unittest.TestCase):
    """Test config updating."""

    def setUp(self):
        self.config = Config(
            root_path="",
            defaults={"kerko": {"one": 1, "two": 2}},
        )

    def test_update_defaults_no_overlap(self):
        config_update(self.config, {"kerko": {"ten": 10, "twenty": 20, "thirty": 30}})
        self.assertEqual(
            dict(self.config),
            {"kerko": {"one": 1, "two": 2, "ten": 10, "twenty": 20, "thirty": 30}},
        )

    def test_update_defaults_with_overlap(self):
        config_update(self.config, {"kerko": {"two": 20, "thirty": 30}})
        self.assertEqual(dict(self.config), {"kerko": {"one": 1, "two": 20, "thirty": 30}})


class ConfigGetSetTestCase(unittest.TestCase):
    """Test config access."""

    def setUp(self):
        self.config = Config(
            root_path="",
            defaults={"kerko": {"one": 1, "none": None, "list": ["a", "b", "c"]}},
        )

    def test_get_existing(self):
        self.assertEqual(config_get(self.config, "kerko.one"), 1)
        self.assertEqual(config_get(self.config, "kerko.none"), None)
        self.assertEqual(config_get(self.config, "kerko.list"), ["a", "b", "c"])

    def test_get_nonexisting(self):
        with self.assertRaises(KeyError):
            config_get(self.config, "kerko.foo")

    def test_set_new_value(self):
        config_set(self.config, "kerko.foo", "bar")
        self.assertEqual(
            dict(self.config),
            {"kerko": {"one": 1, "none": None, "list": ["a", "b", "c"], "foo": "bar"}},
        )

    def test_set_replace_value(self):
        config_set(self.config, "kerko.one", 10)
        self.assertEqual(
            dict(self.config),
            {"kerko": {"one": 10, "none": None, "list": ["a", "b", "c"]}},
        )

    def test_set_replace_list(self):
        config_set(self.config, "kerko.list", ["x", "y", "z"])
        self.assertEqual(
            dict(self.config),
            {"kerko": {"one": 1, "none": None, "list": ["x", "y", "z"]}},
        )


class ParseKerkoConfigTestCase(unittest.TestCase):
    """Test parsing of the kerko config table."""

    def test_default_config(self):
        config = Config(root_path="", defaults=KERKO_DEFAULTS)
        parse_config(config, "kerko", KerkoModel)

    def test_invalid_config(self):
        config = Config(root_path="", defaults={"kerko": "bad_config"})
        with self.assertRaises(RuntimeError):
            parse_config(config, "kerko", KerkoModel)


class ParseRootConfigTestCase(unittest.TestCase):
    """Test parsing of the config root."""

    def test_valid_config(self):
        config = Config(
            root_path="",
            defaults={
                "SECRET_KEY": "1234567890AB",
                "ZOTERO_API_KEY": "1234567890ABCDEFGHIJKLMN",
                "ZOTERO_LIBRARY_ID": "123456",
                "ZOTERO_LIBRARY_TYPE": "group",
                "ignored1": "1",
                "ignored2": 2,
                "ignored3": [1, 2, "three"],
            },
        )
        parse_config(config, None, ConfigModel)
        self.assertEqual(config["ignored1"], "1")
        self.assertEqual(config["ignored2"], 2)
        self.assertEqual(config["ignored3"], [1, 2, "three"])

    def test_coerced_variable_type(self):
        config = Config(
            root_path="",
            defaults={
                "SECRET_KEY": "1234567890AB",
                "ZOTERO_API_KEY": "1234567890ABCDEFGHIJKLMN",
                "ZOTERO_LIBRARY_ID": 123456,  # Integer.
                "ZOTERO_LIBRARY_TYPE": "group",
            },
        )
        parse_config(config, None, ConfigModel)
        self.assertEqual(config["ZOTERO_LIBRARY_ID"], "123456")

    def test_missing_variable(self):
        config = Config(
            root_path="",
            defaults={
                "SECRET_KEY": "1234567890AB",
                "ZOTERO_LIBRARY_ID": "123456",
                "ZOTERO_LIBRARY_TYPE": "group",
                # Missing 'ZOTERO_API_KEY'.
            },
        )
        with self.assertRaises(RuntimeError):
            parse_config(config, None, ConfigModel)

    def test_empty_variable(self):
        config = Config(
            root_path="",
            defaults={
                "SECRET_KEY": "",  # Empty.
                "ZOTERO_API_KEY": "1234567890ABCDEFGHIJKLMN",
                "ZOTERO_LIBRARY_ID": "123456",
                "ZOTERO_LIBRARY_TYPE": "group",
            },
        )
        with self.assertRaises(RuntimeError):
            parse_config(config, None, ConfigModel)

    def test_invalid_variable(self):
        config = Config(
            root_path="",
            defaults={
                "SECRET_KEY": "1234567890AB",
                "ZOTERO_API_KEY": "1234567890ABCDEFGHIJKLMN",
                "ZOTERO_LIBRARY_ID": "ABCDEF",  # Invalid.
                "ZOTERO_LIBRARY_TYPE": "group",
            },
        )
        with self.assertRaises(RuntimeError):
            parse_config(config, None, ConfigModel)



from itertools import chain

from babel.numbers import format_decimal
from flask import url_for
from flask_babel import get_locale

from kerko.shortcuts import config


def get_sections(page_num, page_count):
    """Return dict of pager sections, where each section is a list of page numbers."""
    if page_count <= 1:
        return None

    sections = {}
    page_num = min(max(page_num, 1), page_count)
    first = max(1, int(page_num - config("kerko.pagination.pager_links") / 2))
    last = min(page_count + 1, first + config("kerko.pagination.pager_links") + 1)
    first = max(1, last - config("kerko.pagination.pager_links") - 1)

    if page_num > 1:
        sections["first"] = [1]
        sections["previous"] = [max(1, page_num - 1)]

    sections["before"] = list(range(first, page_num))
    sections["current"] = [page_num]
    sections["after"] = list(range(page_num + 1, last))

    if page_num < page_count:
        sections["next"] = [min(page_count, page_num + 1)]
        sections["last"] = [page_count]

    return sections


def get_page_numbers(sections):
    """Return distinct pages numbers used across pager sections."""
    if sections:
        return sorted(set(chain(*sections.values())))
    return []


def build_pager(sections, criteria, endpoint="kerko.search", **kwargs):
    """
    Build pager links for use in a search view.

    :param str endpoint: The endpoint to pass to `url_for()`.

    :param dict kwargs: Other parameters to pass to `url_for()`.
    """
    pager = {}
    if sections:
        for section, pages in sections.items():
            pager[section] = [
                {
                    "page_num": p,
                    "page_num_formatted": format_decimal(p, locale=get_locale()),
                    "url": url_for(
                        endpoint,
                        **criteria.params(
                            options={
                                "page": p if p > 1 else None,
                                "id": None,
                            }
                        ),
                        **kwargs,
                    ),
                }
                for p in pages
            ]
    return pager



import pathlib
from abc import ABC, abstractmethod
from datetime import date, datetime, time
from decimal import Decimal
from typing import Annotated, Any, Optional, Union

import dpath
import whoosh
from flask import Config
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    NonNegativeInt,
    RootModel,
    ValidationError,
    field_validator,
    model_validator,
)
from typing_extensions import Literal

from kerko.specs import (
    LinkByEndpointSpec,
    LinkByURLSpec,
    LinkGroupSpec,
    LinkSpec,
    PageLinkSpec,
    PageSpec,
)

try:
    import tomllib
except ModuleNotFoundError:
    import tomli as tomllib


# Note: To preserve field ordering in Pydantic models, we annotate an
# attribute's type even when it could be determined by its default value.
# See https://docs.pydantic.dev/latest/usage/models/#field-ordering

SlugStr = Annotated[str, Field(pattern=r"^[a-z][a-z0-9_\-]*$")]
URLPathStr = Annotated[str, Field(pattern=r"^/[a-z0-9_\-/]*$")]
FieldNameStr = Annotated[str, Field(pattern=r"^[a-z][a-zA-Z0-9_]*$")]
ElementIdStr = Annotated[str, Field(pattern=r"^[a-z][a-zA-Z0-9]*$")]
IdentifierStr = Annotated[str, Field(pattern=r"^[a-z][a-z0-9_]*$")]
ZoteroItemIdStr = Annotated[str, Field(pattern=r"^[A-Z0-9]{8}$")]


class AssetsModel(BaseModel):
    """Model for the kerko.assets config table."""

    model_config = ConfigDict(extra="forbid")

    bootstrap_version: str
    jquery_version: str
    popper_version: str
    with_jquery: bool
    with_popper: bool


class FeaturesModel(BaseModel):
    """Model for the kerko.features config table."""

    model_config = ConfigDict(extra="forbid")

    download_attachment_new_window: bool
    download_item: bool
    download_results: bool
    download_results_max_count: NonNegativeInt
    open_in_zotero_app: bool
    open_in_zotero_web: bool
    print_item: bool
    print_results: bool
    print_results_max_count: NonNegativeInt
    relations_links: bool
    relations_initial_limit: int = Field(ge=5)
    relations_sort: SlugStr
    results_abstracts: bool
    results_abstracts_max_length: NonNegativeInt
    results_abstracts_max_length_leeway: NonNegativeInt
    results_abstracts_toggler: bool
    results_attachment_links: bool
    results_url_links: bool


class FeedsModel(BaseModel):
    """Model for the kerko.feeds config table."""

    model_config = ConfigDict(extra="forbid")

    formats: list[Optional[Literal["atom"]]]
    fields: list[FieldNameStr]
    max_days: NonNegativeInt
    require_any: dict[FieldNameStr, list[Union[str, bool, int, float]]]
    reject_any: dict[FieldNameStr, list[Union[str, bool, int, float]]]


class MetaModel(BaseModel):
    """Model for the kerko.meta config table."""

    model_config = ConfigDict(extra="forbid")

    title: str
    highwirepress_tags: bool
    google_analytics_id: str


class PaginationModel(BaseModel):
    """Model for the kerko.pagination config table."""

    model_config = ConfigDict(extra="forbid")

    page_len: int = Field(ge=2)
    pager_links: int = Field(ge=2)


class BreadcrumbModel(BaseModel):
    """Model for the kerko.breadcrumb config table."""

    model_config = ConfigDict(extra="forbid")

    enabled: bool
    include_current: bool
    text_max_length: NonNegativeInt
    text_max_length_leeway: NonNegativeInt


class TemplatesModel(BaseModel):
    """Model for the kerko.templates config table."""

    model_config = ConfigDict(extra="forbid")

    base: str
    layout: str
    search: str
    search_item: str
    item: str
    page: str
    atom_feed: str


class ZoteroModel(BaseModel):
    """Model for the kerko.zotero config table."""

    model_config = ConfigDict(extra="forbid")

    batch_size: int = Field(ge=20)
    max_attempts: int = Field(ge=1)
    wait: int = Field(ge=120)
    csl_style: str
    locale: str = Field(pattern=r"^[a-z]{2,3}-[A-Za-z]+$")
    item_include_re: str
    item_exclude_re: str
    tag_include_re: str
    tag_exclude_re: str
    child_include_re: str
    child_exclude_re: str
    attachment_mime_types: list[str]


class PerformanceModel(BaseModel):
    """Model for the kerko.performance config table."""

    model_config = ConfigDict(extra="forbid")

    whoosh_index_memory_limit: int = Field(ge=16)
    whoosh_index_processors: int = Field(ge=1)


class SearchModel(BaseModel):
    """Model for the kerko.search config table."""

    model_config = ConfigDict(extra="forbid")

    result_fields: list[FieldNameStr]
    fulltext: bool
    whoosh_language: str = Field(pattern=r"^[a-z]{2,3}$")

    @field_validator("whoosh_language")
    @classmethod
    def validate_whoosh_has_language(cls, v):
        if not whoosh.lang.has_stemmer(v):
            msg = f"language '{v}' not supported by Whoosh"
            raise ValueError(msg)
        return v


class ScopesModel(BaseModel):
    """Model for the kerko.scopes config table."""

    model_config = ConfigDict(extra="forbid")

    enabled: bool = True
    selector_label: Optional[str] = None
    breadbox_label: Optional[str] = None
    help_text: Optional[str] = None
    weight: int = 0


class CoreRequiredSearchFieldModel(BaseModel):
    """Model for the kerko.search_fields.core.required config table."""

    model_config = ConfigDict(extra="forbid")

    scopes: list[SlugStr]
    boost: float


class CoreOptionalSearchFieldModel(BaseModel):
    """Model for the kerko.search_fields.core.optional config table."""

    model_config = ConfigDict(extra="forbid")

    enabled: bool = True
    scopes: list[SlugStr]
    boost: float


class ZoteroFieldModel(BaseModel):
    """Model for the kerko.search_fields.zotero config table."""

    model_config = ConfigDict(extra="forbid")

    enabled: bool = True
    scopes: list[SlugStr]
    analyzer: Literal["id", "text", "name"]
    boost: float


class CoreSearchFieldsModel(BaseModel):
    model_config = ConfigDict(extra="forbid")

    required: dict[FieldNameStr, CoreRequiredSearchFieldModel]
    optional: dict[FieldNameStr, CoreOptionalSearchFieldModel]


class SearchFieldsModel(BaseModel):
    """Base model for the kerko.search_fields config table."""

    model_config = ConfigDict(extra="forbid")

    core: CoreSearchFieldsModel
    zotero: dict[FieldNameStr, ZoteroFieldModel]


class BaseFacetModel(BaseModel, ABC):
    """Base model for the kerko.facets config table."""

    model_config = ConfigDict(extra="forbid")

    enabled: bool = True
    filter_key: SlugStr
    weight: int = 0
    initial_limit: NonNegativeInt = 0
    initial_limit_leeway: NonNegativeInt = 2
    sort_by: list[Literal["label", "count"]] = ["label"]
    sort_reverse: bool = False
    item_view: bool = True


class TagFacetModel(BaseFacetModel):
    type: Literal["tag"]
    title: Optional[str] = None


class ItemTypeFacetModel(BaseFacetModel):
    type: Literal["item_type"]
    title: Optional[str] = None
    item_view: bool = False


class YearFacetModel(BaseFacetModel):
    type: Literal["year"]
    title: Optional[str] = None
    item_view: bool = False


class LanguageFacetModel(BaseFacetModel):
    type: Literal["language"]
    title: Optional[str] = None
    item_view: bool = False
    values_separator_re: str = Field(";", min_length=1)
    normalize: bool = True
    locale: str = Field("en", pattern=r"^[a-z]{2,3}(-[A-Za-z]+)?$")
    allow_invalid: bool = False


class LinkFacetModel(BaseFacetModel):
    type: Literal["link"]
    title: Optional[str] = None
    item_view: bool = False


class CollectionFacetModel(BaseFacetModel):
    type: Literal["collection"]
    title: str
    collection_key: str = Field(pattern=r"^[A-Z0-9]{8}$")


# Note: Discriminated unions ensure that a single unambiguous error gets
# reported when validation fails. Reference:
# https://docs.pydantic.dev/latest/usage/types/#discriminated-unions-aka-tagged-unions

FacetModelUnion = Annotated[
    Union[
        TagFacetModel,
        ItemTypeFacetModel,
        YearFacetModel,
        LanguageFacetModel,
        LinkFacetModel,
        CollectionFacetModel,
    ],
    Field(discriminator="type"),
]


class SortsModel(BaseModel):
    """Model for the kerko.sorts config table."""

    model_config = ConfigDict(extra="forbid")

    enabled: bool = True
    weight: int = 0
    label: Optional[str] = None


class BibFormatsModel(BaseModel):
    """Model for the kerko.bib_formats config table."""

    model_config = ConfigDict(extra="forbid")

    enabled: bool = True
    weight: int = 0
    label: Optional[str] = None
    help_text: Optional[str] = None
    extension: SlugStr
    mime_type: str


class RelationsModel(BaseModel):
    """Model for the kerko.relations config table."""

    model_config = ConfigDict(extra="forbid")

    enabled: bool = True
    weight: int = 0
    label: Optional[str] = None


class PageModel(BaseModel):
    """Model for items under the kerko.pages config table."""

    model_config = ConfigDict(extra="forbid")

    path: URLPathStr
    item_id: ZoteroItemIdStr
    title: str

    def to_spec(self) -> PageSpec:
        return PageSpec(
            path=self.path,
            item_id=self.item_id,
            title=self.title,
        )


class PagesModel(RootModel):
    """
    Model for the kerko.pages config table.

    The dictionary key will serve as the page's endpoint name, hence the
    constrained `IdentifierStr` type.
    """

    root: dict[IdentifierStr, PageModel]

    def to_spec(self) -> dict[str, PageSpec]:
        return {key: page_model.to_spec() for key, page_model in self.root.items()}


class LinkModel(BaseModel, ABC):
    model_config = ConfigDict(extra="forbid")

    text: str
    weight: int = 0
    new_window: bool = False

    @abstractmethod
    def to_spec(self) -> LinkSpec:
        pass


class LinkByEndpointModel(LinkModel):
    """Model for endpoint items under the kerko.link_groups config table."""

    type: Literal["endpoint"]
    endpoint: str
    anchor: Optional[str] = None
    scheme: Optional[str] = None
    external: bool = False
    parameters: Optional[dict[str, Any]] = None

    @model_validator(mode="after")
    def validate_scheme(self):
        if self.scheme and not self.external:
            msg = "When specifying 'scheme', 'external' must be true."
            raise ValueError(msg)
        return self

    def to_spec(self) -> LinkSpec:
        return LinkByEndpointSpec(
            text=self.text,
            weight=self.weight,
            new_window=self.new_window,
            endpoint=self.endpoint,
            anchor=self.anchor,
            scheme=self.scheme,
            external=self.external,
            parameters=self.parameters,
        )


class LinkByURLModel(LinkModel):
    """Model for URL items under the kerko.link_groups config table."""

    type: Literal["url"]
    url: str  # TODO: Validate as URL string

    def to_spec(self) -> LinkSpec:
        return LinkByURLSpec(
            text=self.text,
            weight=self.weight,
            new_window=self.new_window,
            url=self.url,
        )


class PageLinkModel(LinkModel):
    """Model for page items under the kerko.link_groups config table."""

    type: Literal["page"]
    page: str

    def to_spec(self) -> LinkSpec:
        return PageLinkSpec(
            text=self.text,
            weight=self.weight,
            page=self.page,
        )


LinkModelUnion = Annotated[
    Union[
        LinkByEndpointModel,
        LinkByURLModel,
        PageLinkModel,
    ],
    Field(discriminator="type"),
]


class LinkGroupsModel(RootModel):
    root: dict[SlugStr, Annotated[list[LinkModelUnion], Field(min_items=1)]]

    def to_spec(self) -> dict[str, LinkGroupSpec]:
        return {
            key: LinkGroupSpec(key, [link_model.to_spec() for link_model in links])
            for key, links in self.root.items()
        }


class KerkoModel(BaseModel):
    """Model for the kerko config table."""

    model_config = ConfigDict(extra="forbid")

    assets: AssetsModel
    features: FeaturesModel
    feeds: FeedsModel
    meta: MetaModel
    pagination: PaginationModel
    breadcrumb: BreadcrumbModel
    pages: Optional[PagesModel] = None
    link_groups: LinkGroupsModel
    templates: TemplatesModel
    zotero: ZoteroModel
    performance: PerformanceModel
    search: SearchModel
    scopes: dict[SlugStr, ScopesModel]
    search_fields: SearchFieldsModel
    facets: dict[FieldNameStr, FacetModelUnion]
    sorts: dict[SlugStr, SortsModel]
    bib_formats: dict[SlugStr, BibFormatsModel]
    relations: dict[ElementIdStr, RelationsModel]


class ConfigModel(BaseModel):
    """Model for validating mandatory at the root config table."""

    # Note: This model allows extra fields since we cannot cover all variables
    # that Flask, Flask extensions, and applications might support.

    model_config = ConfigDict(coerce_numbers_to_str=True)

    SECRET_KEY: str = Field(min_length=12)
    ZOTERO_API_KEY: str = Field(min_length=16)
    ZOTERO_LIBRARY_ID: str = Field(pattern=r"^[0-9]+$")
    ZOTERO_LIBRARY_TYPE: Literal["user", "group"]
    kerko: Optional[KerkoModel] = None


def load_toml(filename: Union[str, pathlib.Path], verbose=False) -> dict[str, Any]:
    """Load the content of a TOML file."""
    try:
        with pathlib.Path(filename).open("rb") as file:
            config = tomllib.load(file)
            if verbose:
                print(f" * Loading configuration file {filename}")  # noqa: T201
            return config
    except OSError as e:
        msg = f"Unable to open TOML file.\n{e}"
        raise RuntimeError(msg) from e
    except tomllib.TOMLDecodeError as e:
        msg = f"Invalid TOML format in file '{filename}'.\n{e}"
        raise RuntimeError(msg) from e


def config_update(config: Config, new_data: dict[str, Any]) -> None:
    """
    Update the configuration with the specified `dict`.

    Unlike the standard `dict.update`, this performs a deep merge. However, it
    does not deep copy the objects.
    """
    dpath.merge(config, new_data, flags=dpath.MergeType.REPLACE)


def config_get(config: Config, path: str) -> Any:
    """
    Retrieve an arbitrarily nested configuration parameter.

    `path` is a string of keys separated by dots ('.') acting as hierarchical
    level separators.

    Unlike `dict.get()`, this provides no way of specifying a default value as
    all configuration parameters are expected to have be already initialized
    with a default value. Passing an invalid `path`, i.e., trying to access a
    parameter that does not have a value, is considered a programming error and
    will throw a `KeyError` exception.
    """
    return dpath.get(config, path, separator=".")


def config_set(config: Config, path: str, value: Any) -> None:
    """
    Set an arbitrarily nested configuration parameter.

    `path` is a string of keys separated by dots ('.') acting as hierarchical
    level separators.
    """
    dpath.new(config, path, value, separator=".")


def parse_config(
    config: Config,
    key: Optional[str] = None,
    model: type[BaseModel] = ConfigModel,
) -> None:
    """
    Parse and validate configuration using `model`.

    Configuration values in the `config` object get replaced by validated ones.
    Values may thus get silently coerced into the types specified by the model
    (unless strict typing is enforced by the model, in which case an error will
    be raised).

    This function also inserts a parsed representation of the configuration
    under key 'kerko_config'.

    If `key` is `None`, then the whole configuration gets parsed with the given
    `model`. Otherwise only the structure at the specified `key` gets parsed.

    If `key` is not `None` but is absent from the config, parsing is silently
    skipped.
    """
    try:
        if key is None:
            parsed = model.model_validate(config)
            config.update(parsed.model_dump())
            config["kerko_config"] = parsed
        elif config.get(key):
            # The parsed models are stored in the config as dicts. This way, the
            # whole configuration structure is made of dicts only, allowing
            # consistent access for any element at any depth.
            parsed = model.model_validate(config[key])
            config_set(config, key, parsed.model_dump())
            config[f"kerko_config.{key}"] = parsed
    except ValidationError as e:
        msg = f"Invalid configuration. {e}"
        raise RuntimeError(msg) from e


def is_toml_serializable(obj: object) -> bool:
    """
    Check if the given object would be serializable into a TOML file.

    This only performs a shallow check of the object.
    """
    return isinstance(
        obj, (bool, int, float, date, datetime, time, Decimal, str, list, tuple, dict)
    )



import time
from datetime import datetime

from babel.numbers import format_decimal
from flask import redirect, render_template, url_for
from flask_babel import get_locale, gettext, ngettext
from werkzeug.datastructures import MultiDict

from kerko.criteria import create_feed_criteria
from kerko.searcher import Searcher
from kerko.shortcuts import composer, config
from kerko.storage import load_object, open_index
from kerko.views import breadbox, pager, sorter
from kerko.views.item import build_item_context, inject_item_data


def _build_item_search_urls(items, criteria):
    """Generate an URL for each search result (switching the search to page-len=1)."""

    def build_url(item, position):
        return url_for(
            ".search",
            **criteria.params(
                options={
                    "page": (criteria.options.get("page", 1) - 1) * page_len + position + 1,
                    "page-len": 1,
                    "id": item["id"],
                }
            ),
        )

    page_len = criteria.options.get("page-len", config("kerko.pagination.page_len"))
    if page_len == "all":
        page_len = 0
    return [build_url(item, position) for position, item in enumerate(items)]


def empty_results(criteria, form):
    """Prepare the template context variables for an empty search results page."""
    context = {}
    facets = {}
    context["title"] = gettext("Your search did not match any resources")
    context["total_count"] = context["total_count_formatted"] = 0
    context["page_count"] = context["page_count_formatted"] = 0
    if criteria.has_filters():
        # When search results are empty, Whoosh cannot provide any groupings,
        # thus our facets cannot be built. However, we still want to display the
        # breadbox with active filters, if any. To build that, we perform a
        # separate search query for each active facet, each time ignoring all
        # other search criteria. Unless a given facet value alone leads to empty
        # results, we'll be able to get to build that facet.
        index = open_index("index")
        with Searcher(index) as searcher:
            for key, value in criteria.filters.lists():
                results = searcher.search(
                    filters=MultiDict({key: value}),
                    reject_any={"item_type": ["note", "attachment"]},
                    limit=1,  # We don't care about the actual items.
                    faceting=True,
                )
                facets.update(results.facets(composer().facets, criteria, active_only=True))
    context["breadbox"] = breadbox.build_breadbox(criteria, facets)
    return render_template(
        config("kerko.templates.search"),
        form=form,
        locale=get_locale(),
        is_searching=criteria.is_searching(),
        **context,
    )


def search_single(criteria, form):
    """Perform search, and prepare template context for a results page containing a single item."""
    start_time = time.process_time()
    context = {}
    index = open_index("index")
    with Searcher(index) as searcher:
        results = searcher.search_page(
            page=criteria.options.get("page", 1),
            page_len=1,
            keywords=criteria.keywords,
            filters=criteria.filters,
            reject_any={"item_type": ["note", "attachment"]},
            sort_spec=criteria.get_active_sort_spec(),
            faceting=True,
        )

        criteria_id = criteria.options.get("id")
        if criteria_id and (results.is_empty() or criteria_id != results[0]["id"]):
            # The search URL is obsolete, the result no longer matches the expected item.
            return redirect(url_for(".item_view", item_id=criteria_id, _external=True), 301)

        if results.is_empty():
            return empty_results(criteria, form)

        criteria.fit_page(results.page_count)
        if criteria.is_searching():
            context["search_title"] = gettext("Your search")
        else:
            context["search_title"] = gettext("Full bibliography")

        if not criteria_id:
            # Item id is not present in the URL. Inject it into the preferred URL, which will allow
            # this view to redirect to the item page, should the same URL be accessed later but the
            # search results be different.
            context["preferred_url"] = url_for(
                ".search",
                **criteria.params(
                    options={
                        "id": results[0]["id"],
                    },
                ),
            )

        # Load item with all available fields.
        item = results.items(composer().fields, composer().facets)[0]
        facets = results.facets(composer().facets, criteria, active_only=True)

        inject_item_data(item)
        context.update(build_item_context(item))
        context["total_count"] = results.item_count
        context["total_count_formatted"] = format_decimal(results.item_count, locale=get_locale())
        context["page_count"] = results.page_count
        context["page_count_formatted"] = format_decimal(results.page_count, locale=get_locale())
        pager_sections = pager.get_sections(criteria.options["page"], results.page_count)
        context["pager"] = pager.build_pager(pager_sections, criteria)
        context["breadbox"] = breadbox.build_breadbox(criteria, facets)
        context["back_url"] = url_for(
            ".search",
            **criteria.params(
                options={
                    "page": int(
                        (criteria.options.get("page", 1) - 1) / config("kerko.pagination.page_len")
                        + 1
                    ),
                    "page-len": None,
                    "id": None,
                }
            ),
        )
    return render_template(
        config("kerko.templates.search_item"),
        form=form,
        time=time.process_time() - start_time,
        locale=get_locale(),
        is_searching=criteria.is_searching(),
        **context,
    )


def search_list(criteria, form):
    """Perform search, and prepare the template context variables for a list of search results."""
    start_time = time.process_time()
    context = {}
    index = open_index("index")
    with Searcher(index) as searcher:
        page_len = criteria.options.get("page-len", config("kerko.pagination.page_len"))
        common_search_args = {
            "keywords": criteria.keywords,
            "filters": criteria.filters,
            "reject_any": {"item_type": ["note", "attachment"]},
            "sort_spec": criteria.get_active_sort_spec(),
            "faceting": True,
        }
        if page_len == "all":
            results = searcher.search(
                limit=None,
                **common_search_args,
            )
        else:
            results = searcher.search_page(
                page=criteria.options.get("page", 1),
                page_len=page_len,
                **common_search_args,
            )

        if results.is_empty():
            return empty_results(criteria, form)

        criteria.fit_page(results.page_count)

        if criteria.is_searching():
            context["title"] = ngettext("Result", "Results", results.item_count)
        else:
            context["title"] = gettext("Full bibliography")
        context["total_count"] = results.item_count
        context["total_count_formatted"] = format_decimal(results.item_count, locale=get_locale())
        context["page_count"] = results.page_count
        context["page_count_formatted"] = format_decimal(results.page_count, locale=get_locale())
        context["pager"] = pager.build_pager(
            pager.get_sections(criteria.options["page"], results.page_count),
            criteria,
        )
        context["sorter"] = sorter.build_sorter(criteria)
        context["show_abstracts"] = criteria.options.get(
            "abstracts", config("kerko.features.results_abstracts")
        )
        context["abstracts_toggler_url"] = url_for(
            ".search",
            **criteria.params(
                options={
                    "abstracts": int(
                        not criteria.options.get(
                            "abstracts", config("kerko.features.results_abstracts")
                        )
                    ),
                }
            ),
        )
        context["print_url"] = url_for(
            ".search",
            **criteria.params(
                options={
                    "page": None,
                    "page-len": "all",
                    "print-preview": 1,
                }
            ),
        )
        context["print_preview"] = criteria.options.get("print-preview", 0)
        context["download_urls"] = {
            key: url_for(
                ".search_bib_download",
                bib_format_key=key,
                **criteria.params(
                    options={
                        "page": None,
                    }
                ),
            )
            for key in composer().bib_formats.keys()
        }
        if "atom" in config("kerko.feeds.formats"):
            context["atom_feed_url"] = url_for(
                ".atom_feed",
                **create_feed_criteria(initial=criteria).params(
                    options={
                        "page": None,
                    }
                ),
            )
            if criteria.is_searching():
                context["atom_feed_help"] = gettext("Custom feed based on your search")
                context["atom_feed_title"] = gettext("Custom feed")
            else:
                context["atom_feed_title"] = gettext("Main feed")

        # Prepare search result items.
        field_specs = composer().select_fields(
            config("kerko.search.result_fields")
            + [badge.field.key for badge in composer().badges.values()],
        )
        items = results.items(field_specs)
        results_facets = results.facets(composer().facets, criteria)
        context["search_results"] = zip(items, _build_item_search_urls(items, criteria))
        context["facet_results"] = results_facets
        context["breadbox"] = breadbox.build_breadbox(criteria, results_facets)

        last_sync = load_object("index", "last_update_from_zotero")
        if last_sync:
            context["last_sync"] = datetime.fromtimestamp(
                last_sync,
                tz=datetime.now().astimezone().tzinfo,
            )
    return render_template(
        config("kerko.templates.search"),
        form=form,
        time=time.process_time() - start_time,
        locale=get_locale(),
        is_searching=criteria.is_searching(),
        **context,
    )



"""
Unit tests for the datetime module.
"""

import unittest
from datetime import datetime

from flask import Flask, current_app
from flask_babel import Babel

import kerko
from kerko.datetime import iso_to_datetime, maximize_partial_date, parse_partial_date, reformat_date


class ParsePartialDateTestCase(unittest.TestCase):
    """Test the `parse_partial_date()` function."""

    def test_ymd(self):
        y, m, d = parse_partial_date("2020-11-17")
        self.assertEqual(y, 2020)
        self.assertEqual(m, 11)
        self.assertEqual(d, 17)

    def test_ym(self):
        y, m, d = parse_partial_date("2020-11", default_day=5)
        self.assertEqual(y, 2020)
        self.assertEqual(m, 11)
        self.assertEqual(d, 5)

    def test_y(self):
        y, m, d = parse_partial_date("2020", default_month=12, default_day=5)
        self.assertEqual(y, 2020)
        self.assertEqual(m, 12)
        self.assertEqual(d, 5)

    def test_no_validation_expected(self):
        y, m, d = parse_partial_date("9999-13-32")
        self.assertEqual(y, 9999)
        self.assertEqual(m, 13)
        self.assertEqual(d, 32)

    def test_partially_parsable_date(self):
        y, m, d = parse_partial_date("1987-MM-DD")
        self.assertEqual(y, 1987)
        self.assertEqual(m, 0)
        self.assertEqual(d, 0)
        self.assertIsInstance(m, int)
        self.assertIsInstance(d, int)

    def test_no_date(self):
        y, m, d = parse_partial_date("no-date-here")
        self.assertEqual(y, 0)
        self.assertEqual(m, 0)
        self.assertEqual(d, 0)
        self.assertIsInstance(y, int)
        self.assertIsInstance(m, int)
        self.assertIsInstance(d, int)

    def test_empty_input(self):
        y, m, d = parse_partial_date("")
        self.assertEqual(y, 0)
        self.assertEqual(m, 0)
        self.assertEqual(d, 0)
        self.assertIsInstance(y, int)
        self.assertIsInstance(m, int)
        self.assertIsInstance(d, int)

    def test_wrong_input_type(self):
        with self.assertRaises(TypeError):
            parse_partial_date(False)

    def test_none_input_type(self):
        with self.assertRaises(TypeError):
            parse_partial_date(None)


class MaximizeDateTestCase(unittest.TestCase):
    """Test the `maximize_partial_date()` function."""

    def test_ymd_date(self):
        y, m, d = maximize_partial_date(1991, 8, 24)
        self.assertEqual(y, 1991)
        self.assertEqual(m, 8)
        self.assertEqual(d, 24)

    def test_invalid_ymd_date(self):
        # The function is not meant to validate dates, so these just pass through.
        y, m, d = maximize_partial_date(1000, 2000, 3000)
        self.assertEqual(y, 1000)
        self.assertEqual(m, 2000)
        self.assertEqual(d, 3000)

    def test_zero_date(self):
        today = datetime.today()
        y, m, d = maximize_partial_date(0, 0, 0)
        self.assertEqual(y, today.year)
        self.assertEqual(m, today.month)
        self.assertEqual(d, today.day)

    def test_ym_date(self):
        y, m, d = maximize_partial_date(2014, 2, 0)
        self.assertEqual(y, 2014)
        self.assertEqual(m, 2)
        self.assertEqual(d, 28)

    def test_y_date(self):
        y, m, d = maximize_partial_date(1996, 0, 0)
        self.assertEqual(y, 1996)
        self.assertEqual(m, 12)
        self.assertEqual(d, 31)

    def test_this_y_ymd_date(self):
        today = datetime.today()
        y, m, d = maximize_partial_date(today.year, 7, 1)
        self.assertEqual(y, today.year)
        self.assertEqual(m, 7)
        self.assertEqual(d, 1)

    def test_this_ym_ymd_date(self):
        today = datetime.today()
        y, m, d = maximize_partial_date(today.year, today.month, 24)
        self.assertEqual(y, today.year)
        self.assertEqual(m, today.month)
        self.assertEqual(d, 24)

    def test_this_ym_ym_date(self):
        today = datetime.today()
        y, m, d = maximize_partial_date(today.year, today.month, 0)
        self.assertEqual(y, today.year)
        self.assertEqual(m, today.month)
        self.assertEqual(d, today.day)

    def test_this_y_y_date(self):
        today = datetime.today()
        y, m, d = maximize_partial_date(today.year, 0, 0)
        self.assertEqual(y, today.year)
        self.assertEqual(m, today.month)
        self.assertEqual(d, today.day)

    def test_future_ymd_date(self):
        y, m, d = maximize_partial_date(2305, 7, 13)
        self.assertEqual(y, 2305)
        self.assertEqual(m, 7)
        self.assertEqual(d, 13)

    def test_future_y_date(self):
        y, m, d = maximize_partial_date(2332, 0, 0)
        self.assertEqual(y, 2332)
        self.assertEqual(m, 12)
        self.assertEqual(d, 31)


class DateStringReformattingTestCase(unittest.TestCase):
    """Test the date string reformatting function."""

    def setUp(self):
        self.app = Flask(__name__)
        self.app.register_blueprint(kerko.make_blueprint(), url_prefix="/bibliography")
        self.babel = Babel()
        ctx = self.app.app_context()
        ctx.push()

    def test_non_iso8601_utc_en(self):
        current_app.config["BABEL_DEFAULT_LOCALE"] = "en_US"
        current_app.config["BABEL_DEFAULT_TIMEZONE"] = "UTC"
        self.babel.init_app(current_app)
        self.assertEqual(reformat_date("2020"), "2020")
        self.assertEqual(reformat_date("2020-10"), "2020-10")
        self.assertEqual(reformat_date("2020-05-05"), "2020-05-05")
        self.assertEqual(reformat_date("01/02/2021"), "01/02/2021")
        self.assertEqual(reformat_date("September 2020"), "September 2020")
        self.assertEqual(reformat_date("September 20, 2020"), "September 20, 2020")
        self.assertEqual(reformat_date("September 20, 2020", convert_tz=True), "September 20, 2020")
        self.assertEqual(
            reformat_date("September 20, 2020", convert_tz=True, show_tz=True), "September 20, 2020"
        )

    def test_iso8601_utc_en(self):
        current_app.config["BABEL_DEFAULT_LOCALE"] = "en_US"
        current_app.config["BABEL_DEFAULT_TIMEZONE"] = "UTC"
        self.babel.init_app(current_app)
        self.assertRegex(
            reformat_date("2020-09-17T07:11:41+00:00", convert_tz=True), r"9/17/20, 7:11\sAM"
        )
        self.assertRegex(
            reformat_date("2020-09-17T07:11:41+00:00", convert_tz=False), r"9/17/20, 7:11\sAM"
        )
        self.assertRegex(
            reformat_date("2020-09-17T07:11:41+00:00", convert_tz=True, show_tz=True),
            r"9/17/20, 7:11\sAM\s\(UTC\)",
        )
        self.assertRegex(
            reformat_date("2020-11-07T12:30:19Z", convert_tz=True), r"11/7/20, 12:30\sPM"
        )
        self.assertRegex(
            reformat_date("2020-11-07T12:30:19Z", convert_tz=False), r"11/7/20, 12:30\sPM"
        )
        self.assertRegex(
            reformat_date("2020-11-07T12:30:19Z", convert_tz=True, show_tz=True),
            r"11/7/20, 12:30\sPM\s\(UTC\)",
        )
        self.assertRegex(
            reformat_date("2019-07-22T20:42+00:00", convert_tz=True), r"7/22/19, 8:42\sPM"
        )
        self.assertRegex(
            reformat_date("2019-07-22T20:42+00:00", convert_tz=False), r"7/22/19, 8:42\sPM"
        )
        self.assertRegex(
            reformat_date("2019-07-22T20:42+00:00", convert_tz=True, show_tz=True),
            r"7/22/19, 8:42\sPM\s\(UTC\)",
        )
        self.assertRegex(
            reformat_date("2020-11-07T12:30:19-0500", convert_tz=True), r"11/7/20, 5:30\sPM"
        )
        self.assertRegex(
            reformat_date("2020-11-07T12:30:19-0500", convert_tz=False), r"11/7/20, 12:30\sPM"
        )
        self.assertRegex(
            reformat_date("2020-11-07T12:30:19-0500", convert_tz=True, show_tz=True),
            r"11/7/20, 5:30\sPM\s\(UTC\)",
        )

    def test_iso8601_est_en(self):
        current_app.config["BABEL_DEFAULT_LOCALE"] = "en_US"
        current_app.config["BABEL_DEFAULT_TIMEZONE"] = "EST"
        self.babel.init_app(current_app)
        self.assertRegex(
            reformat_date("2020-09-17T07:11:41+00:00", convert_tz=True), r"9/17/20, 2:11\sAM"
        )
        self.assertRegex(
            reformat_date("2020-09-17T07:11:41+00:00", convert_tz=False), r"9/17/20, 7:11\sAM"
        )
        self.assertRegex(
            reformat_date("2020-09-17T07:11:41+00:00", convert_tz=True, show_tz=True),
            r"9/17/20, 2:11\sAM\s\(EST\)",
        )
        self.assertRegex(
            reformat_date("2020-11-07T12:30:19Z", convert_tz=True), r"11/7/20, 7:30\sAM"
        )
        self.assertRegex(
            reformat_date("2020-11-07T12:30:19Z", convert_tz=False), r"11/7/20, 12:30\sPM"
        )
        self.assertRegex(
            reformat_date("2020-11-07T12:30:19Z", convert_tz=True, show_tz=True),
            r"11/7/20, 7:30\sAM\s\(EST\)",
        )
        self.assertRegex(
            reformat_date("2019-07-22T20:42+00:00", convert_tz=True), r"7/22/19, 3:42\sPM"
        )
        self.assertRegex(
            reformat_date("2019-07-22T20:42+00:00", convert_tz=False), r"7/22/19, 8:42\sPM"
        )
        self.assertRegex(
            reformat_date("2019-07-22T20:42+00:00", convert_tz=True, show_tz=True),
            r"7/22/19, 3:42\sPM\s\(EST\)",
        )
        self.assertRegex(
            reformat_date("2020-11-07T12:30:19-0500", convert_tz=True), r"11/7/20, 12:30\sPM"
        )
        self.assertRegex(
            reformat_date("2020-11-07T12:30:19-0500", convert_tz=False), r"11/7/20, 12:30\sPM"
        )
        self.assertRegex(
            reformat_date("2020-11-07T12:30:19-0500", convert_tz=True, show_tz=True),
            r"11/7/20, 12:30\sPM\s\(EST\)",
        )

    def test_non_iso8601_utc_fr(self):
        current_app.config["BABEL_DEFAULT_LOCALE"] = "fr_FR"
        current_app.config["BABEL_DEFAULT_TIMEZONE"] = "UTC"
        self.babel.init_app(current_app)
        self.assertEqual(reformat_date("September 2020"), "September 2020")
        self.assertEqual(reformat_date("September 20, 2020"), "September 20, 2020")

    def test_iso8601_utc_fr(self):
        current_app.config["BABEL_DEFAULT_LOCALE"] = "fr_FR"
        current_app.config["BABEL_DEFAULT_TIMEZONE"] = "UTC"
        self.babel.init_app(current_app)
        self.assertEqual(
            reformat_date("2020-09-17T07:11:41+00:00", convert_tz=True), "17/09/2020 07:11"
        )
        self.assertEqual(
            reformat_date("2020-09-17T07:11:41+00:00", convert_tz=False), "17/09/2020 07:11"
        )
        self.assertEqual(
            reformat_date("2020-09-17T07:11:41+00:00", convert_tz=True, show_tz=True),
            "17/09/2020 07:11 (UTC)",
        )
        self.assertEqual(reformat_date("2020-11-07T12:30:19Z", convert_tz=True), "07/11/2020 12:30")
        self.assertEqual(
            reformat_date("2020-11-07T12:30:19Z", convert_tz=False), "07/11/2020 12:30"
        )
        self.assertEqual(
            reformat_date("2020-11-07T12:30:19Z", convert_tz=True, show_tz=True),
            "07/11/2020 12:30 (UTC)",
        )
        self.assertEqual(
            reformat_date("2019-07-22T20:42+00:00", convert_tz=True), "22/07/2019 20:42"
        )
        self.assertEqual(
            reformat_date("2019-07-22T20:42+00:00", convert_tz=False), "22/07/2019 20:42"
        )
        self.assertEqual(
            reformat_date("2019-07-22T20:42+00:00", convert_tz=True, show_tz=True),
            "22/07/2019 20:42 (UTC)",
        )
        self.assertEqual(
            reformat_date("2020-11-07T12:30:19-0500", convert_tz=True), "07/11/2020 17:30"
        )
        self.assertEqual(
            reformat_date("2020-11-07T12:30:19-0500", convert_tz=False), "07/11/2020 12:30"
        )
        self.assertEqual(
            reformat_date("2020-11-07T12:30:19-0500", convert_tz=True, show_tz=True),
            "07/11/2020 17:30 (UTC)",
        )


class IsoToDatetimeTestCase(unittest.TestCase):
    """Test parsing ISO 8601 date strings into datetime objects."""

    def test_date(self):
        dt = iso_to_datetime("1988-08-18T20:38:12Z")
        self.assertEqual(dt.year, 1988)
        self.assertEqual(dt.month, 8)
        self.assertEqual(dt.day, 18)
        self.assertEqual(dt.hour, 20)
        self.assertEqual(dt.minute, 38)
        self.assertEqual(dt.second, 12)

    def test_unsupported_date_format(self):
        with self.assertRaises(ValueError):
            iso_to_datetime("1988-08-18")

    def test_non_existing_date(self):
        with self.assertRaises(ValueError):
            iso_to_datetime("1988-04-31T20:38:12Z")

    def test_invalid_input(self):
        with self.assertRaises(ValueError):
            iso_to_datetime("not-a-date")

    def test_wrong_input_type(self):
        with self.assertRaises(TypeError):
            iso_to_datetime(False)

    def test_none_input_type(self):
        with self.assertRaises(TypeError):
            iso_to_datetime(None)



from collections import defaultdict


class Tree(defaultdict):
    """A dict-based tree structure."""

    # Inspiration for this class: https://gist.github.com/hrldcpr/2012250.
    # However, this implements it as a class to isolate the caller from
    # the implementation, and to include a conversion method.

    def __init__(self):
        # Pass the class itself as the factory.
        super().__init__(Tree)

    def to_dict(self):
        """Convert to regular dict (prettier when printed)."""

        def convert(x):
            try:
                return x.to_dict()
            except AttributeError:
                return x

        return {k: convert(self[k]) for k in self}


if __name__ == "__main__":
    import pprint

    dummy = Tree()
    dummy["some"]["dummy"]["path"] = "data"
    pprint.pprint(dummy.to_dict())  # noqa: T203
